from tqdm import tqdm
import itertools
from ..cost import input, output
from ..openai_calls import openai_call
from typing import List

# K is a constant factor that determines how much ratings change
K = 32

class Elo:
    def __init__(self, test_cases: List[str], prompts: List[str], description: str, model_test: str='gpt-3.5-turbo', model_test_temperature: float=0.6, model_test_max_tokens: int=1000, best_prompts: int=2, timeout: int=10, n_retries: int=5):

        """
        Initialize an Elo instance.

        Args:
            description (str): Description of the use-case.
            test_cases (list): List of test cases to evaluate.
            number_of_prompts (int): Number of prompts to generate and test.
            model_test (str): The language model used for testing.
            model_test_temperature (float): The temperature parameter for the testing model.
            model_test_max_tokens (int): The maximum number of tokens allowed for the testing model.
            model_generation (str): The language model used for generating prompts.
            model_generation_temperature (float): The temperature parameter for the generation model.
            prompts (list): List of prompts to evaluate.
            best_prompts (int): Number of best prompts to consider.

        Note:
            The 'system_gen_system_prompt' and 'ranking_system_prompt' attribute is predefined within the class constructor.
        """

        self.description = description
        self.test_cases = test_cases
        self.model_test = model_test
        self.model_test_temperature = model_test_temperature
        self.model_test_max_tokens = model_test_max_tokens
        self.system_gen_system_prompt = """Your job is to generate system prompt for GPT, given a description of the use-case and some test cases.

The prompts you will be generating will be for freeform tasks, such as generating a landing page headline, an intro paragraph, solving a math problem, etc.

In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

Most importantly, output NOTHING but the prompt. Do not include anything else in your message. Your resulting prompt must be enclosed in double quotation marks."""

        self.ranking_system_prompt = """Your job is to rank the quality of two outputs generated by different prompts. The prompts are used to generate a response for a given task.

You will be provided with the task description, the test prompt, and two generations - one for each system prompt.

Rank the generations in order of quality. If Generation A is better, respond with 'A'. If Generation B is better, respond with 'B'.

Remember, to be considered 'better', a generation must not just be good, it must be noticeably superior to the other.

Also, keep in mind that you are a very harsh critic. Only rank a generation as better if it truly impresses you more than the other.

Respond with your ranking, and nothing else. Be fair and unbiased in your judgement."""

        self.prompts = prompts
        self.timeout = timeout
        self.n_retries = n_retries
        self.best_prompts = best_prompts

    def expected_score(self, r1, r2):

        """
        Calculate the expected score based on Elo rating.

        Args:
            r1 (float): Elo rating of the first prompt.
            r2 (float): Elo rating of the second prompt.

        Returns:
            float: Expected score.
        """

        return 1 / (1 + 10**((r2 - r1) / 400))

    def update_elo(self, r1, r2, score1):

        """
        Update Elo ratings based on the match outcome.

        Args:
            r1 (float): Elo rating of the first prompt.
            r2 (float): Elo rating of the second prompt.
            score1 (float): Score of the first prompt.

        Returns:
            tuple: Updated Elo ratings for the first and second prompt.
        """

        e1 = self.expected_score(r1, r2)
        e2 = self.expected_score(r2, r1)
        return r1 + K * (score1 - e1), r2 + K * ((1 - score1) - e2)

    def get_score(self, test_case, pos1, pos2):

        """
        Get the score of two prompts.

        Args:
            test_case (str): The test case.
            pos1 (str): Prompt 1.
            pos2 (str): Prompt 2.

        Returns:
            tuple: A tuple containing score, cost, input tokens used, and output tokens used.
        """
        
        model=self.model_test
        messages=[
                {"role": "system", "content": self.ranking_system_prompt},
                {"role": "user", "content": f"""Task: {self.description.strip()}
    Prompt: {test_case}
    Generation A: {pos1}
    Generation B: {pos2}"""}
        ]
        logit_bias={
                '32': 100,  # 'A' token
                '33': 100,  # 'B' token
            }
        max_tokens=1
        temperature=self.model_test_temperature

        score = openai_call.create_chat_completion(model, messages, max_tokens, temperature, 1, logit_bias=logit_bias,timeout=self.timeout, n_retries=self.n_retries)

        tokens_input = score.usage.prompt_tokens
        tokens_output = score.usage.completion_tokens
        cost_input = input.cost(tokens_input, self.model_test)
        cost_output = output.cost(tokens_output, self.model_test)
        cost = cost_input + cost_output
        return score.choices[0].message.content, cost, tokens_input, tokens_output

    def get_generation(self, prompt, test_case):

        """
        Generate a response using a prompt and a test case.

        Args:
            prompt (str): The prompt.
            test_case (str): The test case.

        Returns:
            tuple: A tuple containing generated response, cost, input tokens used, and output tokens used.
        """

        model=self.model_test
        messages=[
            {"role": "system", "content": prompt},
            {"role": "user", "content": f"{test_case}"}
        ]
        max_tokens=self.model_test_max_tokens
        temperature=self.model_test_temperature

        generation = openai_call.create_chat_completion(model, messages, max_tokens, temperature, 1, timeout=self.timeout, n_retries=self.n_retries)

        tokens_input = generation.usage.prompt_tokens
        tokens_output = generation.usage.completion_tokens
        cost_input = input.cost(tokens_input, self.model_test)
        cost_output = output.cost(tokens_output, self.model_test)
        cost = cost_input + cost_output
        return generation.choices[0].message.content, cost, tokens_input, tokens_output

    def test_candidate_prompts(self):

        """
        Test candidate prompts using Elo ratings and generate rankings.

        Returns:
            tuple: A tuple containing prompt ratings, battle results, Elo ratings, cost, input tokens used, and output tokens used.
        """
        
        cost = 0
        tokens_input = 0
        tokens_output = 0

    # Initialize each prompt with an ELO rating of 1200
        prompt_ratings = {}
        i = 0
        while i < len(self.prompts):
            prompt = self.prompts[i]
            prompt_ratings[prompt] = 1200
            i += 1

        # Calculate total rounds for progress bar
        total_rounds = len(self.test_cases) * len(self.prompts) * (len(self.prompts) - 1) // 2

        # Initialize progress bar
        pbar = tqdm(total=total_rounds, ncols=70)

        battles = [{"description": self.description, "method": 'ELO'}]
        elo_prompt = []
        for prompt in self.prompts:
            elo_prompt.append({"prompt": prompt, "elo": 1200})
        # For each pair of prompts
        for prompt1, prompt2 in itertools.combinations(self.prompts, 2):
            # For each test case
            for test_case in self.test_cases:
                # Get the value of the "prompt" field as a string
                prompt_content = test_case

                # Update progress bar
                pbar.update()

                # Generate outputs for each prompt
                generation1 = self.get_generation(prompt1, test_case)
                generation2 = self.get_generation(prompt2, test_case)
                response1 = generation1[0]
                response2 = generation2[0]
                cost_generation1 = generation1[1]
                cost_generation2 = generation2[1]
                tokens_input1 = generation1[2]
                tokens_input2 = generation2[2]
                tokens_output1 = generation1[3]
                tokens_output2 = generation2[3]

                # Rank the outputs
                ranking1 = self.get_score(test_case, response1, response2)
                ranking2 = self.get_score(test_case, response2, response1)
                score1 = ranking1[0]
                score2 = ranking2[0]
                cost_score1 = ranking1[1]
                cost_score2 = ranking2[1]
                tokens_input_rank1 = ranking1[2]
                tokens_input_rank2 = ranking2[2]
                tokens_output_rank1 = ranking1[3]
                tokens_output_rank2 = ranking2[3]

                # Convert scores to numeric values
                score1 = 1 if score1 == 'A' else 0 if score1 == 'B' else 0.5
                score2 = 1 if score2 == 'B' else 0 if score2 == 'A' else 0.5

                # Average the scores
                score = (score1 + score2) / 2

                # Update ELO ratings
                r1, r2 = prompt_ratings[prompt1], prompt_ratings[prompt2]
                r1, r2 = self.update_elo(r1, r2, score)
                elo_prompt.append({"prompt": prompt1, "elo": r1})
                elo_prompt.append({"prompt": prompt2, "elo": r2})
                prompt_ratings[prompt1], prompt_ratings[prompt2] = r1, r2

                # Print the winner of this round and save the results
                if score > 0.5:
                    battles.append({"test": prompt_content, "prompt1": prompt1, "generation1": generation1[0], "prompt2": prompt2, "generation2": generation2[0], "winner": prompt1})
                    print(f"Winner: {prompt1}")
                elif score < 0.5:
                    battles.append({"test": prompt_content, "prompt1": prompt1, "generation1": generation1[0], "prompt2": prompt2, "generation2": generation2[0], "winner": prompt2})
                    print(f"Winner: {prompt2}")
                else:
                    battles.append({"test": prompt_content, "prompt1": prompt1, "generation1": generation1[0], "prompt2": prompt2, "generation2": generation2[0], "winner": 'Draw'})
                    print("Draw")

                cost = cost + cost_generation1 + cost_generation2 + cost_score1 + cost_score2
                tokens_input = tokens_input + tokens_input1 + tokens_input_rank1 + tokens_input2 + tokens_input_rank2
                tokens_output = tokens_output + tokens_output1 + tokens_output_rank1 + tokens_output2 + tokens_output_rank2

        elo_prompt_sorted = sorted(elo_prompt, key=lambda x: x["prompt"])
        # Close progress bar
        pbar.close()
        return prompt_ratings, battles, elo_prompt_sorted, cost, tokens_input, tokens_output

    def evaluate_optimal_prompt(self): 

        """
        Evaluate and determine the optimal prompt based on Elo ratings.

        Returns:
            tuple: A tuple containing data list, best prompts, Elo ratings, cost, input tokens used, and output tokens used.

        This method calculates Elo ratings for a set of prompts, ranks them based on their Elo ratings, and returns
        information about the best prompts and the evaluation results.
        """
        
        # Calculate Elo ratings and gather evaluation results
        
        prompt_ratings = self.test_candidate_prompts()
        # Prepare a data list for prompts and ratings, sorted by rating in descending order
        data_list = []
        for prompt, rating in sorted(prompt_ratings[0].items(), key=lambda item: item[1], reverse=True):
            data_list.append({"prompt": prompt, "rating": rating})
        # Append the battle results and Elo ratings to the data list
        data_list.append(prompt_ratings[1])
        data_list.append(prompt_ratings[2])
        # Select the best prompts based on the specified count
        best_prompts = data_list[:self.best_prompts]
        return data_list, best_prompts, prompt_ratings[3], prompt_ratings[4], prompt_ratings[5]


import argparse
import os
import json
import matplotlib.pyplot as plt
from promptwizard.prompt_generation import generation, iteration
from promptwizard.evals import elovalue, classification, equals, includes, function_calling, code_generation, json_validation, semantic_similarity, logprobs
from promptwizard.approximate_cost import cost
import yaml
import textwrap
import numpy as np
from collections import defaultdict
import dotenv
from promptwizard.validation_yaml import validation
from pathlib import Path
from promptwizard.constants import constants
import openai

def valid_yaml(file_name):
    
    # initialize validation
    valid = True

    # Open the YAML file and load its content
    with open(file_name, 'r') as file:
        content = yaml.safe_load(file)

    # Check if 'method' is defined in 'test' section
    if 'method' not in content['test']:
        valid = False
        print("Validation error:")
        print({'test': {'method': ['Must be defined']}}) # Notify about the missing method
    
    # Handle prompts related information
    if 'generation' in content['prompts']:
        if 'number' in content['prompts']['generation']:
            number_of_prompts = content['prompts']['generation']['number']
        if not 'number' in content['prompts']['generation']:
            number_of_prompts = 4
    if 'list' in content['prompts']:
        number_of_prompts = len(content['prompts']['list'])

    # Handle prompts iteration and best_prompts validation
    if 'iterations' in content['prompts']:
        if 'best_prompts' in content['prompts']['iterations']:
            best_prompts = content['prompts']['iterations']['best_prompts']
        if not 'best_prompts' in content['prompts']['iterations']:
            best_prompts = 2
        if best_prompts < 2 or best_prompts >= number_of_prompts:
            valid = False
            print("Validation error:")
            print({'prompts': {'iterations': {'best_prompts':['best_prompts has to be greater than or equal to 2 and strictly less than number_of_prompts.']}}})
            return valid
    # Allowed test method names
    allowed_names = ['Function Calling', 'Classification', 'Equals', 'Includes', 'Elo', 'Code Generation', 'JSON Validation', 'Semantic Similarity', 'LogProbs']

    # Check if the selected 'method' is valid
    if 'method' in content['test']:
        if content['test']['method'] == 'Function Calling':
            config_schema = validation.ValidationFunctionCalling()

        if content['test']['method'] == 'Classification' or content['test']['method'] == 'Equals' or content['test']['method'] == 'Includes':
            config_schema = validation.ValidationClaEqIn()

        if content['test']['method'] == 'Elo':
            config_schema = validation.ValidationElo()

        if content['test']['method'] == 'Code Generation':
            config_schema = validation.ValidationCode()

        if content['test']['method'] == 'JSON Validation':
            config_schema = validation.ValidationJSON()

        if content['test']['method'] == 'Semantic Similarity':
            config_schema = validation.ValidationEmbeddings()

        if content['test']['method'] == 'LogProbs':
            config_schema = validation.ValidationLogProbs()
        
        # Check if the selected method is in the allowed names
        if content['test']['method'] not in allowed_names:
            valid = False
            error_message = f"Must be one of the following: {', '.join(allowed_names)}"
            print("Validation error:")
            print({'test': {'method': [error_message]}})

        # Validate content using the appropriate schema
        if content['test']['method'] in allowed_names:
            errors = config_schema.validate(content)
            if not errors:
                validated_object = config_schema.load(content)
                print("Successful validation. Validated object:")
                print(validated_object)
            if errors:
                print("Validation errors:")
                print(errors)
            if errors != {}:
                valid = False

    return valid

def read_yaml(file_name):
    """
    It loads and reads the content of a given YAML file and returns its content as a Python dictionary or list.
    :type file_name: path of the yaml file.
    :return: yaml file as an object.
    """
    with open(file_name, 'r') as file:
        content = yaml.safe_load(file)
    return content

def approximate_cost(file):
    # Extract the absolute path of the YAML file from the input 'file' object
    yaml_file_path = os.path.abspath(file)
    # Read the content of the YAML file using the 'read_yaml' function
    yaml_content = read_yaml(yaml_file_path)

    # Extract essential information from the YAML content
    method = yaml_content['test']['method']
    if method == 'Elo':
        description = yaml_content['test']['description']

    test_cases = yaml_content.get('test', {}).get('cases', [])
    if method == 'Classification' or method == 'Includes' or method == 'Equals' or method == 'JSON Validation' or method == 'Semantic Similarity' or method == 'LogProbs':
        input_output_pairs = [(case['input'], case['output']) for case in test_cases]
        test_cases = input_output_pairs
    if method == 'Function Calling':
        result_list = [[case['input'], case['output1'], case['output2']] for case in test_cases]
        test_cases = result_list
    if method == 'Code Generation':
        result_list = [[case['input'], case['arguments'], case['output']] for case in test_cases]
        test_cases = result_list

    if method == 'Semantic Similarity':
        if 'embeddings' in yaml_content['test']:
            model_embedding = yaml_content['test']['embeddings']['model_name']
        if not 'embeddings' in yaml_content['test']:
            model_embedding = "text-embedding-ada-002"
    if not method == 'Semantic Similarity':
        model_embedding = None

    if not 'model' in yaml_content['test'] and method != "LogProbs":
        model_test = 'gpt-3.5-turbo'
        model_test_max_tokens = 1000
    if not 'model' in yaml_content['test'] and method == "LogProbs":
        model_test = 'gpt-3.5-turbo-instruct'
        model_test_max_tokens = 4
    
    if 'model' in yaml_content['test'] and 'max_tokens' not in yaml_content['test']['model'] and ('name' in yaml_content['test']['model']) and method != "LogProbs":
        model_test_max_tokens = 1000
        model_test = yaml_content['test']['model']['name']
    if 'model' in yaml_content['test'] and 'max_tokens' not in yaml_content['test']['model'] and method == "LogProbs":
        model_test_max_tokens = 4
        model_test = yaml_content['test']['model']['name']
    if 'model' in yaml_content['test'] and 'max_tokens' in yaml_content['test']['model'] and ('name' not in yaml_content['test']['model']) and method != "LogProbs":
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test = 'gpt-3.5-turbo'
    if 'model' in yaml_content['test'] and 'max_tokens' in yaml_content['test']['model'] and ('name' not in yaml_content['test']['model']) and method == "LogProbs":
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test = 'gpt-3.5-turbo-instruct'
        
    if 'model' in yaml_content['test'] and 'name' in yaml_content['test']['model'] and 'max_tokens' in yaml_content['test']['model']:
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])

    if 'list' in yaml_content['prompts']:
        prompts_value = yaml_content['prompts']['list']
        number_of_prompts = len(prompts_value)
    if not 'list' in yaml_content['prompts']:
        prompts_value = []

    if 'generation' in yaml_content['prompts']:

        if 'number' in yaml_content['prompts']['generation']:
            number_of_prompts = int(yaml_content['prompts']['generation']['number'])
        if not 'number' in yaml_content['prompts']['generation']:
            number_of_prompts = 4

        if 'constraints' in yaml_content['prompts']['generation']:
            prompt_constrainst = yaml_content['prompts']['generation']['constraints']
        if not 'constraints' in yaml_content['prompts']['generation']:
            prompt_constrainst = 'None'

        if 'name' not in yaml_content['prompts']['generation']['model'] and ('max_tokens' not in yaml_content['prompts']['generation']['model']):
            model_generation = 'gpt-4-turbo'
            model_generation_max_tokens = 300

        if 'name' in yaml_content['prompts']['generation']['model'] and ('max_tokens' not in yaml_content['prompts']['generation']['model']):
            model_generation = yaml_content['prompts']['generation']['model']['name']
            model_generation_max_tokens = 300

        if 'name' in yaml_content['prompts']['generation']['model'] and ('max_tokens' in yaml_content['prompts']['generation']['model']):
            model_generation = yaml_content['prompts']['generation']['model']['name']
            model_generation_max_tokens = int(yaml_content['prompts']['generation']['model']['max_tokens'])

        if not 'name' in yaml_content['prompts']['generation']['model'] and ('max_tokens' in yaml_content['prompts']['generation']['model']):
            model_generation = 'gpt-4-turbo'
            model_generation_max_tokens = int(yaml_content['prompts']['generation']['model']['max_tokens'])

        if method != 'Elo':
            description = yaml_content['prompts']['generation']['description']

    if not 'generation' in yaml_content['prompts']:
        model_generation = 'gpt-4-turbo'
        model_generation_max_tokens = 0
        prompt_constrainst = 'None'
        description = None

    if 'functions' in yaml_content['test']:
        functions = yaml_content['test']['functions']

    if 'iterations' in yaml_content['prompts']:
        iterations = int(yaml_content['prompts']['iterations']['number'])

        if 'model' in yaml_content['prompts']['iterations']:
            if 'name' in yaml_content['prompts']['iterations']['model'] and ('max_tokens' in yaml_content['prompts']['iterations']['model']):
                model_iteration = yaml_content['prompts']['iterations']['model']['name']
                model_iteration_max_tokens = yaml_content['prompts']['iterations']['model']['max_tokens']
            if 'name' not in yaml_content['prompts']['iterations']['model'] and ('max_tokens' in yaml_content['prompts']['iterations']['model']):
                model_iteration = model_generation
                model_iteration_max_tokens = yaml_content['prompts']['iterations']['model']['max_tokens']
            if 'name' in yaml_content['prompts']['iterations']['model'] and ('max_tokens' not in yaml_content['prompts']['iterations']['model']):
                model_iteration = yaml_content['prompts']['iterations']['model']['name']
                model_iteration_max_tokens = model_generation_max_tokens

        if not 'model' in yaml_content['prompts']['iterations']:
            model_iteration = model_generation
            model_iteration_max_tokens = model_generation_max_tokens

        if 'best_prompts' in yaml_content['prompts']['iterations']:
            best_prompts = yaml_content['prompts']['iterations']['best_prompts']
        if not 'best_prompts' in yaml_content['prompts']['iterations']:
            best_prompts = 2

    if not 'iterations' in yaml_content['prompts']:
        iterations = 0
        model_iteration = 'None'
        model_iteration_max_tokens = 0
        best_prompts = 2

    # Calculate approximate cost based on the extracted information and the 'cost' module
    if method == 'Function Calling':
        approximate_cost = cost.approximate_cost(test_cases, method, prompts_value, model_test, model_test_max_tokens, number_of_prompts, model_generation, model_generation_max_tokens, iterations, functions, prompt_constrainst, model_iteration, model_iteration_max_tokens, best_prompts, model_embedding)
    if method == 'Elo' or method == 'Classification' or method == 'Equals' or method == 'Includes' or method == 'Code Generation' or method == 'JSON Validation' or method =='LogProbs':
        approximate_cost = cost.approximate_cost(test_cases, method, prompts_value, model_test, model_test_max_tokens, number_of_prompts, model_generation, model_generation_max_tokens, iterations, None, prompt_constrainst, model_iteration, model_iteration_max_tokens, best_prompts, model_embedding)
    
    if method == 'Semantic Similarity':
        approximate_cost = cost.approximate_cost(test_cases, method, prompts_value, model_test, model_test_max_tokens, number_of_prompts, model_generation, model_generation_max_tokens, iterations, None, prompt_constrainst, model_iteration, model_iteration_max_tokens, best_prompts, model_embedding)

    return approximate_cost

def run_evaluation(file, approximate_cost):

    # Extract the absolute path of the YAML file from the input 'file' object
    yaml_file_path = os.path.abspath(file)
    # Read the content of the YAML file using the 'read_yaml' function
    yaml_content = read_yaml(yaml_file_path)

    # Extract essential information from the YAML content
    method = yaml_content['test']['method']
    if method == 'Elo':
        description = yaml_content['test']['description']
    test_cases = yaml_content.get('test', {}).get('cases', [])

    if method == 'Semantic Similarity':
        if 'embeddings' in yaml_content['test']:
            model_embedding = yaml_content['test']['embeddings']['model_name']
        if not 'embeddings' in yaml_content['test']:
            model_embedding = "text-embedding-ada-002"
            print("model_embedding will be 'text-embedding-ada-002'.")
    
    if not 'model' in yaml_content['test'] and method != 'LogProbs':
        model_test = 'gpt-3.5-turbo'
        model_test_max_tokens = 1000
        model_test_temperature = 0.6
        print("model_test will be gpt-3.5-turbo \nmax_tokens will be 1000 \ntemperature will be 0.6")

    if not 'model' in yaml_content['test'] and method == 'LogProbs':
        model_test = 'gpt-3.5-turbo-instruct'
        model_test_max_tokens = 4
        model_test_temperature = 0.6
        print("model_test will be gpt-3.5-turbo-instruct \nmax_tokens will be 4 \ntemperature will be 0.6")
    
    if 'model' in yaml_content['test'] and (not 'name' in yaml_content['test']['model'] and method != 'LogProbs') and ('max_tokens' in yaml_content['test']['model']) and ('temperature' in yaml_content['test']['model']):
        model_test = 'gpt-3.5-turbo'
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test_temperature = int(yaml_content['test']['model']['temperature'])
        print("model_test will be gpt-3.5-turbo.")

    if 'model' in yaml_content['test'] and (not 'name' in yaml_content['test']['model'] and method == 'LogProbs') and ('max_tokens' in yaml_content['test']['model']) and ('temperature' in yaml_content['test']['model']):
        model_test = 'gpt-3.5-turbo-instruct'
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test_temperature = int(yaml_content['test']['model']['temperature'])
        print("model_test will be gpt-3.5-turbo-instruct.")
    
    if 'model' in yaml_content['test'] and (not 'max_tokens' in yaml_content['test']['model'] and method != 'LogProbs') and ('name' in yaml_content['test']['model']) and ('temperature' in yaml_content['test']['model']):
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = 1000
        model_test_temperature = int(yaml_content['test']['model']['temperature'])
        print("max_tokens will be 1000.")

    if 'model' in yaml_content['test'] and (not 'max_tokens' in yaml_content['test']['model'] and method == 'LogProbs') and ('name' in yaml_content['test']['model']) and ('temperature' in yaml_content['test']['model']):
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = 4
        model_test_temperature = int(yaml_content['test']['model']['temperature'])
        print("max_tokens will be 4.")

    if 'model' in yaml_content['test'] and (not 'name' in yaml_content['test']['model'] and method != 'LogProbs') and (not 'max_tokens' in yaml_content['test']['model']) and ('temperature' in yaml_content['test']['model']):
        model_test = 'gpt-3.5-turbo'
        model_test_max_tokens = 1000
        model_test_temperature = int(yaml_content['test']['model']['temperature'])
        print("model_test will be gpt-3.5-turbo \nmodel_test_max_tokens will be 1000.")

    if 'model' in yaml_content['test'] and (not 'name' in yaml_content['test']['model'] and method == 'LogProbs') and (not 'max_tokens' in yaml_content['test']['model']) and ('temperature' in yaml_content['test']['model']):
        model_test = 'gpt-3.5-turbo-instruct'
        model_test_max_tokens = 4
        model_test_temperature = int(yaml_content['test']['model']['temperature'])
        print("model_test will be gpt-3.5-turbo-instruct \nmodel_test_max_tokens will be 4.")

    if 'model' in yaml_content['test'] and (not 'name' in yaml_content['test']['model'] and method != 'LogProbs') and ('max_tokens' in yaml_content['test']['model']) and (not 'temperature' in yaml_content['test']['model']):
        model_test = 'gpt-3.5-turbo'
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test_temperature = 0.6
        print("model_test will be gpt-3.5-turbo \nmodel_test_temperature will be 0.6.")

    if 'model' in yaml_content['test'] and (not 'name' in yaml_content['test']['model'] and method == 'LogProbs') and ('max_tokens' in yaml_content['test']['model']) and (not 'temperature' in yaml_content['test']['model']):
        model_test = 'gpt-3.5-turbo-instruct'
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test_temperature = 0.6
        print("model_test will be gpt-3.5-turbo-instruct \nmodel_test_temperature will be 0.6.")

    if 'model' in yaml_content['test'] and ('name' in yaml_content['test']['model'] and method != 'LogProbs') and (not 'max_tokens' in yaml_content['test']['model']) and (not 'temperature' in yaml_content['test']['model']):
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = 1000
        model_test_temperature = 0.6
        print("model_test_max_tokens will be 1000 \nmodel_test_temperature will be 0.6.")

    if 'model' in yaml_content['test'] and ('name' in yaml_content['test']['model'] and method == 'LogProbs') and (not 'max_tokens' in yaml_content['test']['model']) and (not 'temperature' in yaml_content['test']['model']):
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = 4
        model_test_temperature = 0.6
        print("model_test_max_tokens will be 4 \nmodel_test_temperature will be 0.6.")
    
    if 'model' in yaml_content['test'] and (not 'temperature' in yaml_content['test']['model']) and ('name' in yaml_content['test']['model']) and ('max_tokens' in yaml_content['test']['model']):
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test_temperature = 0.6
        print("temperature will be 0.6.")
    
    if 'model' in yaml_content['test'] and 'name' in yaml_content['test']['model'] and 'max_tokens' in yaml_content['test']['model'] and 'temperature' in yaml_content['test']['model']:
        model_test = yaml_content['test']['model']['name']
        model_test_max_tokens = int(yaml_content['test']['model']['max_tokens'])
        model_test_temperature = int(yaml_content['test']['model']['temperature'])

    if 'best_prompts' in yaml_content['prompts']:
        best_prompts = yaml_content['prompts']['best_prompts']
    if not 'best_prompts' in yaml_content['prompts']:
        best_prompts = 2

    if 'list' in yaml_content['prompts']:
        prompts_value = yaml_content['prompts']['list']
        number_of_prompts = len(prompts_value)
    if not 'list' in yaml_content['prompts']:
        prompts_value = []

    if 'generation' in yaml_content['prompts']:

        if 'number' in yaml_content['prompts']['generation']:
            number_of_prompts = int(yaml_content['prompts']['generation']['number'])
        if not 'number' in yaml_content['prompts']['generation']:
            number_of_prompts = 4

        if 'constraints' in yaml_content['prompts']['generation']:
            prompt_constrainst = yaml_content['prompts']['generation']['constraints']
        if not 'constraints' in yaml_content['prompts']['generation']:
            prompt_constrainst = 'None'

        if not 'model' in yaml_content['prompts']['generation']:
            model_generation = 'gpt-4-turbo'
            model_generation_max_tokens = 300
            model_generation_temperature = 1.2
            print("model_generation will be gpt-4-turbo \nmodel_generation_max_tokens will be 300 \nmodel_generation_temperature will be 1.2")

        if 'model' in yaml_content['prompts']['generation'] and (not 'name' in yaml_content['prompts']['generation']['model']) and ('max_tokens' in yaml_content['prompts']['generation']['model']) and ('temperature' in yaml_content['prompts']['generation']['model']):
            model_generation = 'gpt-4-turbo'
            model_generation_max_tokens = int(yaml_content['prompts']['generation']['model']['max_tokens'])
            model_generation_temperature = int(yaml_content['prompts']['generation']['model']['temperature'])
            print("model_generation will be gpt-4-turbo")

        if 'model' in yaml_content['prompts']['generation'] and (not 'max_tokens' in yaml_content['prompts']['generation']['model']) and ('name' in yaml_content['prompts']['generation']['model']) and ('temperature' in yaml_content['prompts']['generation']['model']):
            model_generation = yaml_content['prompts']['generation']['model']['name']
            model_generation_max_tokens = 300
            model_generation_temperature = int(yaml_content['prompts']['generation']['model']['temperature'])
            print("model_generation_max_tokens will be 300")

        if 'model' in yaml_content['prompts']['generation'] and (not 'temperature' in yaml_content['prompts']['generation']['model']) and ('name' in yaml_content['prompts']['generation']['model']) and ('max_tokens' in yaml_content['prompts']['generation']['model']):
            model_generation = yaml_content['prompts']['generation']['model']['name']
            model_generation_max_tokens = int(yaml_content['prompts']['generation']['model']['max_tokens'])
            model_generation_temperature = 1.2
            print("model_generation_temperature will be 1.2")

        if 'model' in yaml_content['prompts']['generation'] and (not 'name' in yaml_content['prompts']['generation']['model']) and (not 'max_tokens' in yaml_content['prompts']['generation']['model']) and ('temperature' in yaml_content['prompts']['generation']['model']):
            model_generation = 'gpt-4-turbo'
            model_generation_max_tokens = 300
            model_generation_temperature = int(yaml_content['prompts']['generation']['model']['temperature'])
            print("model_generation will be gpt-4-turbo \nmodel_generation_max_tokens will be 300")

        if 'model' in yaml_content['prompts']['generation'] and (not 'name' in yaml_content['prompts']['generation']['model']) and ('max_tokens' in yaml_content['prompts']['generation']['model']) and (not 'temperature' in yaml_content['prompts']['generation']['model']):
            model_generation = 'gpt-4-turbo'
            model_generation_max_tokens = int(yaml_content['prompts']['generation']['model']['max_tokens'])
            model_generation_temperature = 1.2
            print("model_generation will be gpt-4-turbo \nmodel_generation_temperature will be 1.2")

        if 'model' in yaml_content['prompts']['generation'] and ('name' in yaml_content['prompts']['generation']['model']) and (not 'max_tokens' in yaml_content['prompts']['generation']['model']) and (not 'temperature' in yaml_content['prompts']['generation']['model']):
            model_generation = yaml_content['prompts']['generation']['model']['name']
            model_generation_max_tokens = 300
            model_generation_temperature = 1.2
            print("model_generation_max_tokens will be 300 \nmodel_generation_temperature will be 1.2")
        
        if 'model' in yaml_content['prompts']['generation'] and ('name' in yaml_content['prompts']['generation']['model']) and ('max_tokens' in yaml_content['prompts']['generation']['model']) and ('temperature' in yaml_content['prompts']['generation']['model']):
            model_generation = yaml_content['prompts']['generation']['model']['name']
            model_generation_max_tokens = int(yaml_content['prompts']['generation']['model']['max_tokens'])
            model_generation_temperature = int(yaml_content['prompts']['generation']['model']['temperature'])

        if method != 'Elo':
            description = yaml_content['prompts']['generation']['description']

    if not 'generation' in yaml_content['prompts'] and (not 'list' in yaml_content['prompts']):
        model_generation = 'gpt-4-turbo'
        model_generation_max_tokens = 0
        prompt_constrainst = None
        if method != 'Elo':
            description = None
        model_generation_temperature = 0

    if 'functions' in yaml_content['test']:
        functions = yaml_content['test']['functions']
        if 'function_call' in yaml_content['test']:
            function_call = yaml_content['test']['function_call']
        else:
            function_call = 'auto'

    if 'iterations' in yaml_content['prompts']:

        iterations = int(yaml_content['prompts']['iterations']['number'])

        if 'model' in yaml_content['prompts']['iterations']:
            if not 'name' in yaml_content['prompts']['iterations']['model'] and ('max_tokens' in yaml_content['prompts']['iterations']['model']) and ('temperature' in yaml_content['prompts']['iterations']['model']):
                model_iteration = model_generation
                model_iteration_max_tokens = yaml_content['prompts']['iterations']['model']['max_tokens']
                model_iteration_temperature = yaml_content['prompts']['iterations']['model']['temperature']

            elif not 'max_tokens' in yaml_content['prompts']['iterations']['model'] and ('name' in yaml_content['prompts']['iterations']['model']) and ('temperature' in yaml_content['prompts']['iterations']['model']):
                model_iteration_max_tokens = model_generation_max_tokens
                model_iteration = yaml_content['prompts']['iterations']['model']['name']
                model_iteration_temperature = yaml_content['prompts']['iterations']['model']['temperature']

            elif not 'temperature' in yaml_content['prompts']['iterations']['model'] and ('name' in yaml_content['prompts']['iterations']['model']) and ('max_tokens' in yaml_content['prompts']['iterations']['model']):
                model_iteration_temperature = model_generation_temperature
                model_iteration = yaml_content['prompts']['iterations']['model']['name']
                model_iteration_max_tokens = yaml_content['prompts']['iterations']['model']['max_tokens']
            
            elif not 'name' in yaml_content['prompts']['iterations']['model'] and (not 'max_tokens' in yaml_content['prompts']['iterations']['model']) and ('temperature' in yaml_content['prompts']['iterations']['model']):
                model_iteration = model_generation
                model_iteration_max_tokens = model_generation_max_tokens
                model_iteration_temperature = yaml_content['prompts']['iterations']['model']['temperature']

            elif not 'name' in yaml_content['prompts']['iterations']['model'] and (not 'temperature' in yaml_content['prompts']['iterations']['model']) and ('max_tokens' in yaml_content['prompts']['iterations']['model']):
                model_iteration = model_generation
                model_iteration_max_tokens = yaml_content['prompts']['iterations']['model']['max_tokens']
                model_iteration_temperature = model_generation_temperature

            elif not 'max_tokens' in yaml_content['prompts']['iterations']['model'] and (not 'temperature' in yaml_content['prompts']['iterations']['model']) and ('name' in yaml_content['prompts']['iterations']['model']):
                model_iteration = yaml_content['prompts']['iterations']['model']['name']
                model_iteration_max_tokens = model_generation_max_tokens
                model_iteration_temperature = model_generation_temperature

            else:
                model_iteration = yaml_content['prompts']['iterations']['model']['name']
                model_iteration_max_tokens = yaml_content['prompts']['iterations']['model']['max_tokens']
                model_iteration_temperature = yaml_content['prompts']['iterations']['model']['temperature']

        if not 'model' in yaml_content['prompts']['iterations']:
            model_iteration = model_generation
            model_iteration_max_tokens = model_generation_max_tokens
            model_iteration_temperature = model_generation_temperature

        if 'best_percentage' in yaml_content['prompts']['iterations']:
            best_percentage = yaml_content['prompts']['iterations']['best_percentage']

        if not ('best_percentage' in yaml_content['prompts']['iterations']) and (method != 'Elo' or method != 'Semantic Similarity'):
            best_percentage = 100
            if method != 'Elo' and method != 'Semantic Similarity' and method != 'LogProbs':
                print("The percentage to be overcome by your best prompts to stop the iteration will be 100%.")

    if not 'iterations' in yaml_content['prompts']:
        iterations = 0
        best_percentage = 100

    if 'timeout' in yaml_content:
        timeout = int(yaml_content['timeout'])
    if not 'timeout' in yaml_content:
        timeout = 10
    if 'n_retries' in yaml_content:
        n_retries = int(yaml_content['n_retries'])
    if not 'n_retries' in yaml_content:
        n_retries = 5

        
    cost = 0
    tokens_input_gpt4 = 0
    tokens_output_gpt4 = 0
    tokens_input_gpt35 = 0
    tokens_output_gpt35 = 0
    tokens_embeddings = 0

    # Determine the class corresponding to the selected method
    if method == 'Elo':
        class_method = elovalue.Elo

    if method == 'Classification':
        class_method = classification.Classification

    if method == 'Equals':
        class_method = equals.Equals

    if method == 'Includes':
        class_method = includes.Includes

    if method == 'Function Calling':
        class_method = function_calling.functionCalling

    if method == 'Code Generation':
        class_method = code_generation.codeGeneration

    if method == 'JSON Validation':
        class_method = json_validation.jsonValidation

    if method == 'Semantic Similarity':
        class_method = semantic_similarity.semanticSimilarity

    if method == 'LogProbs':
        class_method = logprobs.LogProbs

    # Initialize an object of the class obtained from the 'method'
    if method != 'Function Calling' and method != 'Elo' and method != 'Semantic Similarity':
        object_class = class_method(test_cases, None, model_test, model_test_temperature, model_test_max_tokens, best_prompts, timeout, n_retries)
    
    if method == 'Semantic Similarity':
        object_class = class_method(test_cases, None, model_test, model_test_temperature, model_test_max_tokens, model_embedding, best_prompts, timeout, n_retries)

    if method == 'Function Calling':
        object_class = class_method(test_cases, None, functions, model_test, model_test_temperature, model_test_max_tokens, best_prompts, function_call, timeout, n_retries)
    if method == 'Elo':
        object_class = class_method(test_cases, None, description, model_test, model_test_temperature, model_test_max_tokens, best_prompts, timeout, n_retries)

    # Checks if prompts exist, and generates them if necessary
    if prompts_value == []:
        if prompt_constrainst != 'None':
            prompts_generation_cost = generation.generate_candidate_prompts(object_class.system_gen_system_prompt, object_class.test_cases, description, model_generation, model_generation_temperature, model_generation_max_tokens, number_of_prompts, prompt_constrainst, timeout, n_retries)
            prompts_value = prompts_generation_cost[0]
            cost = cost + prompts_generation_cost[1]
            if model_generation == 'gpt-3.5-turbo':
                tokens_input_gpt35 = tokens_input_gpt35 + prompts_generation_cost[2]
                tokens_output_gpt35 = tokens_output_gpt35 + prompts_generation_cost[3]
            elif model_generation == 'gpt-4' or model_generation == 'gpt-4-turbo' or model_generation == 'gpt-4-1106-preview':
                tokens_input_gpt4 = tokens_input_gpt4 + prompts_generation_cost[2]
                tokens_output_gpt4 = tokens_output_gpt4 + prompts_generation_cost[3]
        else:
            prompts_generation_cost = generation.generate_candidate_prompts(object_class.system_gen_system_prompt, object_class.test_cases, description, model_generation, model_generation_temperature, model_generation_max_tokens, number_of_prompts, prompt_constrainst, timeout, n_retries)
            prompts_value = prompts_generation_cost[0]
            cost = cost + prompts_generation_cost[1]
            if model_generation == 'gpt-3.5-turbo':
                tokens_input_gpt35 = tokens_input_gpt35 + prompts_generation_cost[2]
                tokens_output_gpt35 = tokens_output_gpt35 + prompts_generation_cost[3]
            elif model_generation == 'gpt-4' or model_generation == 'gpt-4-turbo' or model_generation == 'gpt-4-1106-preview':
                tokens_input_gpt4 = tokens_input_gpt4 + prompts_generation_cost[2]
                tokens_output_gpt4 = tokens_output_gpt4 + prompts_generation_cost[3]

    
    # Initialize an object for evaluation
    if method == 'Function Calling':
        evaluable_object = class_method(test_cases, prompts_value, functions, model_test, model_test_temperature, model_test_max_tokens, best_prompts, function_call, timeout, n_retries)
    if method == 'Elo':
        evaluable_object = class_method(test_cases, prompts_value, description, model_test, model_test_temperature, model_test_max_tokens, best_prompts, timeout, n_retries)

    if method == 'Semantic Similarity':
        evaluable_object = class_method(test_cases, prompts_value, model_test, model_test_temperature, model_test_max_tokens, model_embedding, best_prompts, timeout, n_retries)

    if method != 'Function Calling' and method != 'Elo' and method != 'Semantic Similarity':
        evaluable_object = class_method(test_cases, prompts_value, model_test, model_test_temperature, model_test_max_tokens, best_prompts, timeout, n_retries)

    # Evaluate the prompts and gather results
    results = evaluable_object.evaluate_optimal_prompt()
    cost = cost + results[2]
    if model_test == 'gpt-4' or model_test == 'gpt-4-1106-preview' or model_test == 'gpt-4-turbo':
        tokens_input_gpt4 = tokens_input_gpt4 + results[3]
        tokens_output_gpt4 = tokens_output_gpt4 + results[4]
    if model_test == 'gpt-3.5-turbo':
        tokens_input_gpt35 = tokens_input_gpt35 + results[3]
        tokens_output_gpt35 = tokens_output_gpt35 + results[4]
    if model_test == 'gpt-3.5-turbo-instruct':
        tokens_input_gpt35 = tokens_input_gpt35 + results[3]
        tokens_output_gpt35 = tokens_output_gpt35 + results[4]
    yaml_folder = os.path.dirname(file)

    if method == 'Semantic Similarity':
        tokens_embeddings = tokens_embeddings + results[5]
    if method == 'Elo':
        # Group "elo" values by prompt using a dictionary
        elos_by_prompt = defaultdict(list)
        for item in results[0][number_of_prompts + 1]:
            prompt = item["prompt"]
            elo = item["elo"]
            elos_by_prompt[prompt].append(elo)

        # Create a scatter plot
        for prompt, elos in elos_by_prompt.items():
            prompt_truncated = textwrap.shorten(prompt, width=20, placeholder="...")
            x = np.arange(1, len(elos) + 1)
            y = np.array(elos)
            x_smooth = np.linspace(x.min(), x.max(), 200)
            y_smooth = np.interp(x_smooth, x, y)
            plt.plot(x_smooth, y_smooth, linewidth=1.5, markersize=6, label=prompt_truncated)
        plt.xlabel('Comparisons')
        plt.ylabel('Elo')
        plt.title('Scatter Plot: Elo by Prompt')
        plt.legend()
        plt.show()
        output_plot_path = os.path.join(yaml_folder, "scatter_plot.png")
        plt.savefig(output_plot_path)
        print(f"Scatter plot saved in: {output_plot_path}")

    # Full path of the output.json file in the same folder as the YAML
    output_json_path = os.path.join(yaml_folder, "output.json")
    # Convert the result to JSON format and save it to the output.json file
    with open(output_json_path, "w") as json_file:
        json.dump(results[0], json_file, indent=4)
    print(f"Result saved in: {output_json_path}")

    # Make iterations if it's necessary
    old_prompts = results[1]
    number_of_iteration = 1
    if method != 'Elo' and method != 'Semantic Similarity':
        tokens_input_gen_iter = 0
        tokens_output_gen_iter = 0
        tokens_input_test_iter = 0
        tokens_output_test_iter = 0
        all_have_rating_percentage = True
        # Iterate through the list of elements and check the 'rating' value.
        for element in old_prompts:
            if element["rating"] < best_percentage:
                all_have_rating_percentage = False
                break  # If an element with a different rating is found, stop the iteration.
        while (iterations > 0 and (not all_have_rating_percentage)):
            print(f"Prompts from iteration number {number_of_iteration}")
            filename = f'output_iteration_{number_of_iteration}.json'
            json_file_path = os.path.join(yaml_folder, filename)
            combine_prompts = []
            if method == 'Function Calling':
                new_results_prompts_cost = iteration.iterations(test_cases, method, old_prompts, number_of_prompts - best_prompts, functions, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, function_call, None, best_prompts, timeout, n_retries)

            if method != 'Function Calling':
                new_results_prompts_cost = iteration.iterations(test_cases, method, old_prompts, number_of_prompts - best_prompts, None, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, None, None, best_prompts, timeout, n_retries)
            new_results = new_results_prompts_cost[0]
            cost = cost + new_results_prompts_cost[1]
            tokens_input_gen_iter = tokens_input_gen_iter + new_results_prompts_cost[2]
            tokens_output_gen_iter = tokens_output_gen_iter + new_results_prompts_cost[4]
            tokens_input_test_iter = tokens_input_test_iter + new_results_prompts_cost[3]
            tokens_output_test_iter = tokens_output_test_iter + new_results_prompts_cost[5]
            with open(json_file_path, 'w') as file:
                json.dump(new_results[0], file, indent=4)
                print(f"Result saved in: {filename}")
            iterations = iterations - 1
            number_of_iteration = number_of_iteration + 1
            combine_prompts.append(old_prompts)
            combine_prompts.append(new_results[1])
            combined_data = [item for sublist in combine_prompts for item in sublist]
            sorted_data = sorted(combined_data, key=lambda x: x['rating'], reverse=True)
            old_prompts = sorted_data[:best_prompts]
            all_have_rating_percentage = True
            for element in old_prompts:
                if element["rating"] < best_percentage:
                    all_have_rating_percentage = False
                    break  # If an element with a different rating is found, stop the iteration.

    if method == 'Semantic Similarity':
        tokens_input_gen_iter = 0
        tokens_output_gen_iter = 0
        tokens_input_test_iter = 0
        tokens_output_test_iter = 0
        tokens_embeddings_iter = 0
        while (iterations > 0):
            filename = f'output_iteration_{number_of_iteration}.json'
            json_file_path = os.path.join(yaml_folder, filename)
            combine_prompts = []

            new_results_prompts_cost = iteration.iterations(test_cases, method, old_prompts, number_of_prompts - best_prompts, None, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, None, None, best_prompts, model_embedding, timeout, n_retries)
            new_results = new_results_prompts_cost[0]
            cost = cost + new_results_prompts_cost[1]
            tokens_input_gen_iter = tokens_input_gen_iter + new_results_prompts_cost[2]
            tokens_output_gen_iter = tokens_output_gen_iter + new_results_prompts_cost[4]
            tokens_input_test_iter = tokens_input_test_iter + new_results_prompts_cost[3]
            tokens_output_test_iter = tokens_output_test_iter + new_results_prompts_cost[5]
            tokens_embeddings_iter = tokens_embeddings_iter + new_results_prompts_cost[6]
            with open(json_file_path, 'w') as file:
                json.dump(new_results[0], file, indent=4)
                print(f"Result saved in: {filename}")
            iterations = iterations - 1
            number_of_iteration = number_of_iteration + 1
            combine_prompts.append(old_prompts)
            combine_prompts.append(new_results[1])
            combined_data = [item for sublist in combine_prompts for item in sublist]
            sorted_data = sorted(combined_data, key=lambda x: x['rating'], reverse=True)
            old_prompts = sorted_data[:best_prompts]
        tokens_embeddings = tokens_embeddings + tokens_embeddings_iter

    if method == 'Elo':
        tokens_input_gen_iter = 0
        tokens_output_gen_iter = 0
        tokens_input_test_iter = 0
        tokens_output_test_iter = 0
        while iterations > 0:
            print(f"Prompts from iteration number {number_of_iteration}")
            prompt_contents = [item['prompt'] for item in old_prompts]
            filename = f'output_iteration_{number_of_iteration}.json'
            json_file_path = os.path.join(yaml_folder, filename)
            combine_prompts = []
            new_results_prompts_cost = iteration.iterations(test_cases, method, prompt_contents, number_of_prompts - best_prompts, None, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, None, description, best_prompts, timeout, n_retries)
            new_results = new_results_prompts_cost[0]
            cost = cost + new_results_prompts_cost[1]
            tokens_input_gen_iter = tokens_input_gen_iter + new_results_prompts_cost[2]
            tokens_output_gen_iter = tokens_output_gen_iter + new_results_prompts_cost[4]
            tokens_input_test_iter = tokens_input_test_iter + new_results_prompts_cost[3]
            tokens_output_test_iter = tokens_output_test_iter + new_results_prompts_cost[5]
            with open(json_file_path, 'w') as file:
                json.dump(new_results[0], file, indent=4)
                print(f"Result saved in: {filename}")

            elos_by_prompt = defaultdict(list)
            for item in new_results[0][number_of_prompts + 1]:
                prompt = item["prompt"]
                elo = item["elo"]
                elos_by_prompt[prompt].append(elo)

            # Create a scatter plot
            for prompt, elos in elos_by_prompt.items():
                prompt_truncated = textwrap.shorten(prompt, width=20, placeholder="...")
                x = np.arange(1, len(elos) + 1)
                y = np.array(elos)
                x_smooth = np.linspace(x.min(), x.max(), 200)
                y_smooth = np.interp(x_smooth, x, y)
                plt.plot(x_smooth, y_smooth, linewidth=1.5, markersize=6, label=prompt_truncated)
            plt.xlabel('Comparisons')
            plt.ylabel('Elo')
            plt.title('Scatter Plot: Elo by Prompt')
            plt.legend()
            plt.show()
            scatter = f'scatter_plot_{number_of_iteration}.png'
            output_plot_path = os.path.join(yaml_folder, scatter)
            plt.savefig(output_plot_path)
            print(f"Scatter plot saved in: {output_plot_path}")

            iterations = iterations - 1
            number_of_iteration = number_of_iteration + 1
            old_prompts = new_results[1]
    # Ends the calculation of consumed tokens and stores this information

    if model_generation == 'gpt-4' or model_generation == 'gpt-4-turbo' or model_generation == 'gpt-4-1106-preview':
        tokens_input_gpt4 = tokens_input_gpt4 + tokens_input_gen_iter
        tokens_output_gpt4 = + tokens_output_gpt4 + tokens_output_gen_iter
    if model_generation == 'gpt-3.5-turbo':
        tokens_input_gpt35 = tokens_input_gpt35 + tokens_input_gen_iter
        tokens_output_gpt35 = + tokens_output_gpt35 + tokens_output_gen_iter
    if model_test == 'gpt-4' or model_test == 'gpt-4-turbo' or model_test == 'gpt-4-1106-preview':
        tokens_input_gpt4 = tokens_input_gpt4 + tokens_input_test_iter
        tokens_output_gpt4 = + tokens_output_gpt4 + tokens_output_test_iter
    if model_test == 'gpt-3.5-turbo':
        tokens_input_gpt35 = tokens_input_gpt35 + tokens_input_test_iter
        tokens_output_gpt35 = + tokens_output_gpt35 + tokens_output_test_iter
    if model_test == 'gpt-3.5-turbo-instruct':
        tokens_input_gpt35 = tokens_input_gpt35 + tokens_input_test_iter
        tokens_output_gpt35 = + tokens_output_gpt35 + tokens_output_test_iter
    
    filename = f'output_best_prompts_and_results.json'
    json_file_path = os.path.join(yaml_folder, filename)
    if method == 'Semantic Similarity':
        tokens_and_cost = {
        "approximate_cost": approximate_cost,
        "real_cost": cost,
        "tokens_input_gpt-3.5-turbo": tokens_input_gpt35,
        "tokens_output_gpt-3.5-turbo": tokens_output_gpt35,
        "tokens_input_gpt-4": tokens_input_gpt4,
        "tokens_output_gpt-4": tokens_output_gpt4,
        "tokens_embeddings": tokens_embeddings
        }
    if method != 'Semantic Similarity':
        tokens_and_cost = {
        "approximate_cost": approximate_cost,
        "real_cost": cost,
        "tokens_input_gpt-3.5-turbo": tokens_input_gpt35,
        "tokens_output_gpt-3.5-turbo": tokens_output_gpt35,
        "tokens_input_gpt-4": tokens_input_gpt4,
        "tokens_output_gpt-4": tokens_output_gpt4
    }
    old_prompts.append(tokens_and_cost)
    with open(json_file_path, 'w') as file:
        json.dump(old_prompts, file, indent=4)

    print(f"The cost of your evaluation was: {cost} dollars.")
    

def main():
    
    parser = argparse.ArgumentParser(description="Read YAML file and get key values.")
    parser.add_argument("yaml_file", help="Path of the YAML file to read.")
    
    parser.add_argument("--env_path", help="Path to the .env file.", default=None)
    args = parser.parse_args()

    try:
        # Define a function to check if a given path is valid
        def valid_path(path):
            try:
                Path(path)
                return True
            except (ValueError, TypeError):
                return False
        # Check if the provided environment path is valid
        if valid_path(args.env_path):
            # Load environment variables from the specified path
            dotenv.load_dotenv(dotenv_path=args.env_path)
        else:
            # Load environment variables from the default '.env' file in the current working directory
            dotenv.load_dotenv(dotenv_path=os.getcwd()+'/.env')
        if os.getenv(constants.OPENAI_API_KEY):
            openai.api_key = os.getenv(constants.OPENAI_API_KEY)
        else:
            raise Exception("No API key provided into enviroment variables, please configure OPENAI_API_KEY")
        
    except FileNotFoundError:
        # Handle the case where the .env file is not found
        print("Error: The .env file was not found in your directory.")
        exit(1)

    # Check if the OPENAI_API_TYPE environment variable is set to "azure"
    if os.getenv("OPENAI_API_TYPE") == "azure":

        # Check if OPENAI_API_BASE is not defined when using the "azure" API type
        if os.getenv("OPENAI_API_BASE") is None:
            print("Error: OPENAI_API_BASE is required when OPENAI_API_TYPE is 'azure'")
            exit(1)

        # Check if OPENAI_API_VERSION is not defined when using the "azure" API type
        if os.getenv("OPENAI_API_VERSION") is None:
            print("Error: OPENAI_API_VERSION is required when OPENAI_API_TYPE is 'azure'")
            exit(1)
    if os.getenv("OPENAI_API_BASE") != None and os.getenv("OPENAI_API_TYPE") != None and os.getenv("OPENAI_API_VERSION") != None:
        openai_env = {
        "OPENAI_API_BASE": os.getenv("OPENAI_API_BASE"),
        constants.OPENAI_API_KEY: os.getenv(constants.OPENAI_API_KEY),
        "OPENAI_API_TYPE": os.getenv("OPENAI_API_TYPE"),
        "OPENAI_API_VERSION": os.getenv("OPENAI_API_VERSION")
        }
    if not (os.getenv("OPENAI_API_BASE") != None and os.getenv("OPENAI_API_TYPE") != None and os.getenv("OPENAI_API_VERSION") != None):
        openai_env = {
            "OPENAI_API_KEY": os.getenv("OPENAI_API_KEY")
        }

    # Check if the provided YAML file is valid
    if (valid_yaml(args.yaml_file)):
        # Calculate the approximate cost of the evaluation based on the YAML file
        approximate = approximate_cost(args.yaml_file)
        print(f"The cost of your evaluation will be approximately {approximate} dollars.")
        user_input = input("Continue? (Y/N): ").strip().lower()
        if user_input == "y":
            run_evaluation(args.yaml_file, approximate)
        else:
            print("Execution aborted.")

if __name__ == "__main__":
    main()


from .elovalue import *
from .classification import *
from .equals import *
from .includes import *
from .function_calling import *
from .code_generation import *
from .json_validation import *
from .semantic_similarity import *
from .logprobs import *


import tiktoken
from ..cost import input, output, embeddings
from typing import Union, List, Dict

def num_tokens_from_string(string: str, encoding_name: str) -> int:
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.encoding_for_model(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens

def num_tokens_from_messages(messages, model):
    """Return the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
        }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-3.5-turbo" in model:
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
    elif "gpt-3.5-turbo-instruct" in model:
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
    elif "gpt-4" in model:
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    elif "gpt-4-turbo" in model:
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    elif "gpt-4-1106-preview" in model:
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens

def approximate_cost_generation(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_generation, model_generation_max_tokens, iterations):
    tokens_input = 0 # Initialize a variable to count input tokens
    tokens_output = 0 # Initialize a variable to count output tokens

    # Check the method type to determine the system-generated prompt
    if prompts_value == []:

        if method == 'Includes':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Specify in the prompts that you generate that they give a step-by-step response.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
            
        if method == 'Equals':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.
                 
Remember that the prompt should only allow the AI to answer the answer and nothing else. No explanation is necessary.

You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified. I repeat, do not include the test cases.

Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""

        if method == 'Classification':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        The prompts you will be generating will be for classifiers, with 'true' and 'false' being the only possible outputs.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
            
        if method == 'Elo':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

The prompts you will be generating will be for freeform tasks, such as generating a landing page headline, an intro paragraph, solving a math problem, etc.

In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""

        if method == 'Function Calling':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""

        if method == 'Code Generation':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""

        if method == 'JSON Validation':
           system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Clarify in your prompts that you are going to believe that the result of the execution has to be a json object and nothing more, no explanation is necessary.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message.""" 
           
        if method == 'Semantic Similarity':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Specify in the prompts you generate that the response must be text.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
            
        if method == 'LogProbs':
            system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Specify in your prompts that the response has to be with the token with the highest logprobs.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""


        # Create a message list based on the prompt change requirement
        if prompt_change != 'None':
            message = [
                    {"role": "system", "content": system_gen_system_prompt + ' ' + prompt_change},
                    {"role": "user", "content": f"Here are some test cases:`{test_cases}\n\nRespond with your prompt, and nothing else. Be creative."}
                    ]
        if prompt_change == 'None':
            message = [
                    {"role": "system", "content": system_gen_system_prompt},
                    {"role": "user", "content": f"Here are some test cases:`{test_cases}\n\nRespond with your prompt, and nothing else. Be creative."}
                    ]
        tokens_output = tokens_output + number_of_prompts*model_generation_max_tokens
        tokens_input = num_tokens_from_messages(message, model=model_generation)
    cost_input = input.cost(tokens_input, model_generation)
    cost_output = output.cost(tokens_output, model_generation)
    cost = cost_input + cost_output
    return cost

def approximate_cost_test(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_generation, model_generation_max_tokens, iterations, functions, model_embedding):
    tokens_input = 0 # Initialize a variable to count input tokens
    tokens_output = 0 # Initialize a variable to count output tokens
    tokens_embeddings = 0

    # Check the evaluation method to determine token counts
    if method == 'Classification' or method == 'Equals' or method == 'Includes' or method == 'Code Generation' or method == 'JSON Validation' or method == 'LogProbs':
        for test_case in test_cases:
            for i in range(number_of_prompts):
                if method == 'LogProbs' and model_test == 'gpt-3.5-turbo-instruct':
                    model_test = 'gpt-3.5-turbo'
                # Calculate input token count for classification, equals, or includes methods
                tokens_input = tokens_input + 3 + model_generation_max_tokens + num_tokens_from_string(test_case[0], model_test) 
                tokens_output = tokens_output + model_test_max_tokens

    if method == 'Semantic Similarity':
        for test_case in test_cases:
            for i in range(number_of_prompts):
                # Calculate input token count for Semantic Similarity methods
                tokens_input = tokens_input + 3 + model_generation_max_tokens + num_tokens_from_string(test_case[0], model_test) 
                tokens_output = tokens_output + model_test_max_tokens
            if model_embedding != None:
                tokens_embeddings = tokens_embeddings + num_tokens_from_string(test_case[0], model_embedding) 
            

    if method == 'Function Calling':
        tokens_functions = 150*len(functions) # Assuming a constant token count for function descriptions
        for test_case in test_cases:
            for i in range(number_of_prompts):
                # Calculate input token count for Function Calling method
                tokens_input = tokens_input + 3 + model_generation_max_tokens + num_tokens_from_string(test_case[0], model_test) + tokens_functions
                tokens_output = tokens_output + model_test_max_tokens

    if method == 'Elo':
        for test_case in test_cases:
            # Calculate input token count for Elo method
            tokens_input = tokens_input + num_tokens_from_string("""Your job is to rank the quality of two outputs generated by different prompts. The prompts are used to generate a response for a given task.

            You will be provided with the task description, the test prompt, and two generations - one for each system prompt.

            Rank the generations in order of quality. If Generation A is better, respond with 'A'. If Generation B is better, respond with 'B'.

            Remember, to be considered 'better', a generation must not just be good, it must be noticeably superior to the other.

            Also, keep in mind that you are a very harsh critic. Only rank a generation as better if it truly impresses you more than the other.

            Respond with your ranking, and nothing else. Be fair and unbiased in your judgement.""", 'gpt-3.5-turbo') + 2*model_test_max_tokens + 2*num_tokens_from_string(test_case, 'gpt-3.5-turbo') + model_generation_max_tokens
            tokens_output = tokens_output + model_generation_max_tokens + len(test_cases) * len(prompts_value) * (len(prompts_value) - 1) // 2
    
    # Calculate the cost of input tokens using the model's cost function
    cost_input = input.cost(tokens_input, model_test)

    # Calculate the cost of output tokens using the model's cost function
    cost_output = output.cost(tokens_output, model_test)

    # Calculate the cost of embedding tokens using the model's cost function
    cost_embedding = embeddings.cost(tokens_embeddings, model_embedding)

    # Calculate the total cost as the sum of input and output costs
    cost = cost_input + cost_output + cost_embedding

    return cost # Return the calculated cost

def approximate_cost_iterations(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_iteration, model_iteration_max_tokens, iterations, functions, best_prompts):

    if iterations == 0:

        cost = 0

    else:

        # Check the evaluation method to determine the cost calculation strategy
        if method == 'Equals' or method == 'Includes' or method == 'Classification' or method == 'Function Calling' or method == 'Code Generation' or method == 'JSON Validation' or method == 'LogProbs':
            # Calculate the cost for the iterations.
            cost = (approximate_cost_generation(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts -  best_prompts, prompt_change, model_iteration, model_iteration_max_tokens, iterations) + approximate_cost_test(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts - best_prompts, 'None', model_iteration, model_iteration_max_tokens, iterations, functions, None))*iterations

        elif method == 'Semantic Similarity':
            # Calculate the cost for the iterations.
            cost = (approximate_cost_generation(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts -  best_prompts, prompt_change, model_iteration, model_iteration_max_tokens, iterations) + approximate_cost_test(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, 'None', model_iteration, model_iteration_max_tokens, iterations, functions, None))*iterations

        else:
            # Calculate the cost for the iterations
            cost = (approximate_cost_generation(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_iteration, model_iteration_max_tokens, iterations) + approximate_cost_test(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, 'None', model_iteration, model_iteration_max_tokens, iterations, functions, None))*iterations

    return cost # Return the total cost after all iterations

def approximate_cost(test_cases: Union[List[str], List[Dict]], method: str, prompts_value: List[str], model_test: str='gpt-3.5-turbo', model_test_max_tokens: int=1000, number_of_prompts: int=4, model_generation: str='gpt-4', model_generation_max_tokens: int=500, iterations: int=0, functions: List[dict]=None, prompt_change: str=None, model_iteration: str='gpt-4', model_iteration_max_tokens: int=500, best_prompts: int=2, model_embedding: str="text-embedding-ada-002"):
    return approximate_cost_generation(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_generation, model_generation_max_tokens, iterations) + approximate_cost_test(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_generation, model_generation_max_tokens, iterations, functions, model_embedding) + approximate_cost_iterations(test_cases, method, model_test, model_test_max_tokens, prompts_value, number_of_prompts, prompt_change, model_iteration, model_iteration_max_tokens, iterations, functions, best_prompts)


from marshmallow import Schema, fields, ValidationError, validates

class ModelSchema(Schema):
        # Schema for defining the GPT model attributes
        name = fields.Str()
        temperature = fields.Float()
        max_tokens = fields.Integer(strict=True)

        @validates('name')
        def validate_name(self, model):
            # Validate the 'name' field against a list of allowed names
            allowed_names = ['gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'gpt-4-1106-preview']
            if model not in allowed_names:
                raise ValidationError(f"'name' must be one of the following: {', '.join(allowed_names)}")
            
        @validates("temperature")
        def validate_temperature_range(self, value):
            # Validate the 'temperature' field to be within a specified range
            if not (0 <= value <= 2):
                raise ValidationError("'temperature' must be a float between 0 and 2.")
            
        @validates("max_tokens")
        def validate_max_tokens(self, value):
            # Validate the 'max_tokens' field to be greater than 0
            if not (1 <= value):
                raise ValidationError("'max_tokens' must be an integer greater than 0.")
            
class ModelLogProbsSchema(Schema):
    name = fields.Str()
    temperature = fields.Float()
    max_tokens = fields.Integer(strict=True)

    @validates('name')
    def validate_name(self, model):
        # Validate the 'name' field against a list of allowed names
        allowed_names = ['gpt-3.5-turbo-instruct']
        if model != allowed_names[0]:
            raise ValidationError(f"'name' must be one of the following: {', '.join(allowed_names)}")
        
    @validates("temperature")
    def validate_temperature_range(self, value):
        # Validate the 'temperature' field to be within a specified range
        if not (0 <= value <= 2):
            raise ValidationError("'temperature' must be a float between 0 and 2.")
        
    @validates("max_tokens")
    def validate_max_tokens(self, value):
        # Validate the 'max_tokens' field to be greater than 0
        if not (1 <= value <= 4):
            raise ValidationError("'max_tokens' must be an integer greater than 0 and less than or equal to 4.")

class EloFormatTestCaseSchema(Schema):
    # Schema for test cases with the 'Elo' method
    cases = fields.List(fields.Str(required=True))
    method = fields.Str(required=True)
    model = fields.Nested(ModelSchema)
    description = fields.Str(required=True)
        
class CodeTestCaseSchema(Schema):

    input = fields.String(required=True)
    arguments = fields.Raw(required=True)
    output = fields.Raw(required=True)

class ClaEqInTestCaseSchema(Schema):
    # Schema for individual test cases with 'Classification', 'Equals', 'Includes' and 'LogProbs' methods
    input = fields.String(required=True)
    output = fields.String(required=True)

class JSONTestCaseSchema(Schema):

    input = fields.String(required=True)
    output = fields.Dict(required=True)

class EmbeddingsTestCaseSchema(Schema):

    input = fields.String(required=True)
    output = fields.String(required=True)

class ClaEqInFormatTestCaseSchema(Schema):
        # Schema for test cases formatted for 'Classification', 'Equals', and 'Includes' methods
        cases = fields.List(fields.Nested(ClaEqInTestCaseSchema), required=True)
        method = fields.Str(required=True)
        model = fields.Nested(ModelSchema)

        @validates("method")
        def validate_method(self, value):
            # Validate the 'method' field to be one of 'Classification', 'Equals', or 'Includes'
            expected_method1 = "Classification"
            expected_method2 = "Equals"
            expected_method3 = "Includes"
            if value != expected_method1 and value != expected_method2 and value != expected_method3:
                raise ValidationError(f"Method must be '{expected_method1}', '{expected_method2}' '{expected_method3}'.")

class CodeFormatTestCaseSchema(Schema):
        cases = fields.List(fields.Nested(CodeTestCaseSchema), required=True)
        method = fields.Str(required=True)
        model = fields.Nested(ModelSchema)

        @validates("method")
        def validate_method(self, value):
            expected_methods = ['Code Generation']
            if value not in expected_methods:
                raise ValidationError(f"Method must be one of those '{expected_methods}'.")
            
class JSONFormatTestCaseSchema(Schema):
        cases = fields.List(fields.Nested(JSONTestCaseSchema), required=True)
        method = fields.Str(required=True)
        model = fields.Nested(ModelSchema)

class LogProbsFormatTestCaseSchema(Schema):
    # Schema for test cases with the 'LogProbs' method
    cases = fields.List(fields.Nested(ClaEqInTestCaseSchema), required=True)
    method = fields.Str(required=True)
    model = fields.Nested(ModelLogProbsSchema)

    @validates("method")
    def validate_method(self, value):
        # Validate the 'method' field to be 'LogProbs'
        expected_method = "LogProbs"
        if value != expected_method:
            raise ValidationError(f"Method must be '{expected_method}'.")

class EmbeddingsModel(Schema):
     model_name = fields.Str()
     @validates("model_name")
     def validate_method(self, value):
            expected_methods = ['text-embedding-ada-002', 'text-similarity-ada-001', 'text-search-ada-query-001', 'code-search-ada-code-001', 'code-search-ada-text-001', 'text-similarity-babbage-001', 'text-search-babbage-doc-001', 'text-search-babbage-query-001', 'code-search-babbage-code-001', 'code-search-babbage-text-001', 'text-similarity-curie-001', 'text-search-curie-doc-001', 'text-search-curie-query-001', 'text-similarity-davinci-001', 'text-search-davinci-doc-001', 'text-search-davinci-query-001']
            if value not in expected_methods:
                raise ValidationError(f"Method must be one of those '{expected_methods}'.")

class EmbeddingsFormatTestCaseSchema(Schema):
    cases = fields.List(fields.Nested(EmbeddingsTestCaseSchema), required=True)
    method = fields.Str(required=True)
    model = fields.Nested(ModelSchema)
    embeddings = fields.Nested(EmbeddingsModel)

class FunctionCallingFormatFunctionSchema(Schema):
        name = fields.Str(required=True)
        description = fields.Str(required=True)
        parameters = fields.Dict(
        properties=fields.Field(required=True)
        
    )

class FunctionCallingTestCaseSchema(Schema):
    # Schema for defining function attributes used in 'Function Calling' method
    input = fields.Str(required=True)
    output1 = fields.Str(required=True)
    output2 = fields.Str(required=True)

class FunctionCallingFormatTestCaseSchema(Schema):
        # Schema for test cases formatted for 'Function Calling' method
        cases = fields.List(fields.Nested(FunctionCallingTestCaseSchema))
        method = fields.Str(required=True)
        model = fields.Nested(ModelSchema)
        functions = fields.List(fields.Nested(FunctionCallingFormatFunctionSchema))
        function_call = fields.Str()

        @validates("cases")
        def validate_cases_list(self, cases):
            # Validate that each test case consists of 3 strings: test case, correct function, and correct variable
            for case in cases:
                if len(case) != 3:
                    raise ValidationError("Each test case must consist of 3 strings, the first one being the specific test case, the second one the correct function to use, and the third one the correct variable.")
                
        @validates("method")
        def validate_method(self, value):
            # Validate that the 'method' field is 'Function Calling'
            expected_method = "Function Calling"
            if value != expected_method:
                raise ValidationError(f"Method must be '{expected_method}'.")
            
class IterationsSchema(Schema):
        # Schema for defining iteration settings
        number = fields.Integer(strict=True, required=True)
        model = fields.Nested(ModelSchema)
        best_percentage = fields.Float()

        @validates("number")
        def validate_number_iterations(self, value):
            # Validate that 'number' is a non-negative integer
            if not (0 <= value):
                raise ValidationError("'number' must be an integer equal or greater than 0.")
        
        @validates("best_percentage")
        def validate_best_percentage(self, value):
            # Validate that 'best_percentage' is a float between 0 and 100 (inclusive)
            if not (0 < value <= 100):
                raise ValidationError("'best_percentage' must be a float stricter than 0 and less than or equal to 100.")

class GenerationModelEloSchema(Schema):
        # Schema for defining generation settings
        model = fields.Nested(ModelSchema)
        number = fields.Integer(missing=4) # Default number of prompts is 4
        constraints = fields.Str()
        description = fields.Str()

        @validates("number")
        def validate_number_iterations(self, value):
            # Validate that 'number' is an integer equal to or greater than 4
            if not (value >= 4):
                raise ValidationError("'number' must be an integer equal or greater than 4.")

class GenerationModelSchema(Schema):
        # Schema for defining generation settings
        model = fields.Nested(ModelSchema)
        number = fields.Integer(missing=4) # Default number of prompts is 4
        constraints = fields.Str()
        description = fields.Str(required=True)

        @validates("number")
        def validate_number_iterations(self, value):
            # Validate that 'number' is an integer equal to or greater than 4
            if not (value >= 4):
                raise ValidationError("'number' must be an integer equal or greater than 4.")
            
class PromptSchema(Schema):
        # Schema for defining prompts
        list = fields.List(fields.Str())
        iterations = fields.Nested(IterationsSchema)
        generation = fields.Nested(GenerationModelSchema)
        best_prompts = fields.Integer(strict=True)

        @validates("list")
        def validate_my_list_length(self, value):
            # Validate that the prompt list has at least 4 elements
            if value is not None and len(value) < 4:
                raise ValidationError("The list should have at least 4 elements.")

class PromptEloSchema(Schema):
        # Schema for defining prompts
        list = fields.List(fields.Str())
        iterations = fields.Nested(IterationsSchema)
        generation = fields.Nested(GenerationModelEloSchema)

        @validates("list")
        def validate_my_list_length(self, value):
            # Validate that the prompt list has at least 4 elements
            if value is not None and len(value) < 4:
                raise ValidationError("The list should have at least 4 elements.")
        


class ValidationFunctionCalling(Schema):
        # Schema for validating 'Function Calling' test cases and prompts
        test = fields.Nested(FunctionCallingFormatTestCaseSchema, required=True)
        prompts = fields.Nested(PromptSchema, required=True)
        n_retries = fields.Integer()
        timeout = fields.Integer()

class ValidationClaEqIn(Schema):
        # Schema for validating 'Classification', 'Equals', 'Includes' and 'LogProbs' test cases and prompts
        test = fields.Nested(ClaEqInFormatTestCaseSchema, required=True)
        prompts = fields.Nested(PromptSchema, required=True)
        n_retries = fields.Integer()
        timeout = fields.Integer()

class ValidationElo(Schema):
        # Schema for validating 'Elo' test cases and prompts
        test = fields.Nested(EloFormatTestCaseSchema, required=True)
        prompts = fields.Nested(PromptEloSchema, required=True)
        n_retries = fields.Integer()
        timeout = fields.Integer()

class ValidationCode(Schema):
        # Schema for validating 'Code Generation' test cases and prompts
        test = fields.Nested(CodeFormatTestCaseSchema, required=True)
        prompts = fields.Nested(PromptSchema, required=True)
        n_retries = fields.Integer()
        timeout = fields.Integer()

class ValidationJSON(Schema):
    # Schema for validating 'JSON Validation' test cases and prompts
    test = fields.Nested(JSONFormatTestCaseSchema, required=True)
    prompts = fields.Nested(PromptSchema, required=True)
    n_retries = fields.Integer()
    timeout = fields.Integer()

class ValidationEmbeddings(Schema):
    # Schema for validating 'semantic_validation' test cases and prompts
    test = fields.Nested(EmbeddingsFormatTestCaseSchema, required=True)
    prompts = fields.Nested(PromptSchema, required=True)
    n_retries = fields.Integer()
    timeout = fields.Integer()

class ValidationLogProbs(Schema):
    test = fields.Nested(LogProbsFormatTestCaseSchema, required=True)
    prompts = fields.Nested(PromptSchema, required=True)
    n_retries = fields.Integer()
    timeout = fields.Integer()


from ..openai_calls import openai_call
import json
from ..cost import input, output
import concurrent.futures
from typing import List, Dict

class functionCalling:
    def __init__(self, test_cases: List[Dict], prompts: List[str], functions: List[Dict], model_test: str="gpt-3.5-turbo", model_test_temperature: float=0.6, model_test_max_tokens: int=1000, best_prompts: int=2, function_call: str="auto", timeout: int=10, n_retries: int=5):

        """
        Initializes an instance of the functionCalling class.

        Args:
            test_cases (list): A list of test cases, each containing 'input', 'output1', and 'output2' fields.
            number_of_prompts (int): The number of prompts to evaluate.
            model_test (str): The name of the GPT model used for testing.
            model_test_temperature (float): The temperature setting for testing.
            model_test_max_tokens (int): The maximum number of tokens for testing.
            model_generation (str): The name of the GPT model used for prompt generation.
            model_generation_temperature (float): The temperature setting for prompt generation.
            prompts (list): A list of prompts to evaluate.
            functions (list): A list of functions used in function calls.
            function_call (str): The function call format.
            best_prompts (int): The number of best prompts to select.

        Note:
            The 'system_gen_system_prompt' attribute is predefined within the class constructor.
        Initializes the functionCalling class with the provided parameters.
        """

        self.test_cases = test_cases
        self.model_test = model_test
        self.model_test_temperature = model_test_temperature
        self.model_test_max_tokens = model_test_max_tokens
        self.prompts = prompts
        self.best_prompts = best_prompts
        self.functions = functions
        self.function_call = function_call
        self.system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

You should respond a function_call and nothing else. Don't text us back.

You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
        self.timeout = timeout
        self.n_retries = n_retries

    def process_prompt(self, prompt, test_case, model, model_max_tokens, model_temperature, functions, function_call):
        messages = [
            {"role": "system", "content": prompt},
            {"role": "user", "content": f"{test_case['input']}"}
        ]
        response = openai_call.create_chat_completion(model, messages, model_max_tokens, model_temperature, 1, functions=functions, function_call=function_call, timeout=self.timeout, n_retries=self.n_retries)
        partial_tokens_input = response.usage.prompt_tokens
        partial_tokens_output = response.usage.completion_tokens
        print(response)
        response_content = response
        
        return partial_tokens_input, partial_tokens_output, response_content, test_case['output1'], test_case['output2']

    def test_candidate_prompts(self):

        """
        Test candidate prompts against provided test cases.

        Returns:
            tuple: A tuple containing data list, best prompts, cost, input tokens used, and output tokens used.

        This method evaluates a set of prompts by generating responses using a GPT model and comparing the generated
        responses to expected outputs from test cases. It returns information about the performance of each prompt.
        """

        cost = 0
        tokens_input = 0
        tokens_output = 0
        prompt_results = {prompt: {'correct': 0, 'total': 0} for prompt in self.prompts}
        results = [{"method": "Function Calling"}]

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []

            for prompt in self.prompts:
                prompt_and_results = [{"prompt": prompt}]
                for test_case in self.test_cases:
                    future = executor.submit(
                        self.process_prompt,
                        prompt,
                        test_case,
                        self.model_test,
                        self.model_test_max_tokens,
                        self.model_test_temperature,
                        self.functions,
                        self.function_call
                    )
                    futures.append(future)
                    partial_tokens_input, partial_tokens_output, response, ideal_output1, ideal_output2 = future.result()
                    tokens_input += partial_tokens_input
                    tokens_output += partial_tokens_output

                    if "function_call" in response.choices[0].message:
                    # Update model results
                        json_object = json.loads(str(response.choices[0].message.function_call.arguments))

                        def extract_keys_and_values(json_list):
                            all_keys = list(json_object.keys())
                            key_value_pairs = [json_object[key]for key in all_keys]
                            return all_keys, key_value_pairs[0]

                        if ideal_output1 == response.choices[0].message.function_call.name and extract_keys_and_values(json_object)[1] == ideal_output2:
                            prompt_results[prompt]['correct'] += 1
                        prompt_results[prompt]['total'] += 1
                        if ideal_output1 == response.choices[0].message.function_call.name and ideal_output2 == extract_keys_and_values(json_object)[1]:
                            result = True
                        if ideal_output1 == response.choices[0].message.function_call.name and not (ideal_output2 == extract_keys_and_values(json_object)[1]):
                            result = 'variable error'
                        if not (ideal_output1 == response.choices[0].message.function_call.name) and ideal_output2 == extract_keys_and_values(json_object)[1]:
                            result = 'function error'
                        if not (ideal_output1 == response.choices[0].message.function_call.name) and (not ideal_output2 == extract_keys_and_values(json_object)[1]):
                            result = 'function and variable error'
                        prompt_and_results.append({"test": test_case['input'], "answer": {"function": f"{response.choices[0].message.function_call.name}", "variable": f"{extract_keys_and_values(json_object)[1]}"}, "ideal": {"function": f"{ideal_output1}", "variable": f"{ideal_output2}"}, "result": result})
                    if "function_call" not in response.choices[0].message:
                        prompt_and_results.append({"test": test_case['input'], "answer": 'not a function call', "ideal": {"function": f"{ideal_output1}", "variable": f"{ideal_output2}"}, "result": 'Received text data instead of JSON.'})
                        prompt_results[prompt]['total'] += 1

                results.append(prompt_and_results)
                prompt_and_results = []

        cost_input = input.cost(tokens_input, self.model_test)
        cost_output = output.cost(tokens_output, self.model_test)
        cost = cost + cost_input + cost_output

        # Calculate and print the percentage of correct answers and average time for each model
        best_prompt = self.prompts[0]
        best_percentage = 0
        data_list = []
        for i, prompt in enumerate(self.prompts):
            correct = prompt_results[prompt]['correct']
            total = prompt_results[prompt]['total']
            percentage = (correct / total) * 100
            data_list.append({"prompt": prompt, "rating": percentage})
            print(f"Prompt {i+1} got {percentage:.2f}% correct.")
            if percentage >= best_percentage:
                best_percentage = percentage
                best_prompt = prompt
        sorted_data = sorted(data_list, key=lambda x: x['rating'], reverse=True)
        best_prompts = sorted_data[:self.best_prompts]
        print(f"The best prompt was '{best_prompt}' with a correctness of {best_percentage:.2f}%.")
        sorted_data.append(results)
        return sorted_data, best_prompts, cost, tokens_input, tokens_output
    
    def evaluate_optimal_prompt(self):

        """
        Evaluate the optimal prompt for function calling.

        Returns:
            tuple: A tuple containing data list, best prompts, cost, input tokens used, and output tokens used.

        This method evaluates the optimal prompt for function calling by calling the `test_candidate_prompts` method.
        """
        
        return self.test_candidate_prompts()


from ..openai_calls import openai_call
from ..cost import input, output, embeddings
import concurrent.futures
from sklearn.metrics.pairwise import cosine_similarity
import numpy as np
import threading
from typing import List, Dict

class semanticSimilarity:
    def __init__(self, test_cases: List[Dict], prompts: List[str], model_test: str="gpt-3.5-turbo", model_test_temperature: float=0.6, model_test_max_tokens: int=1000, model_embedding: str="text-embedding-ada-002", best_prompts: int=2, timeout: int=10, n_retries: int=5):

        """
        Initialize a semantic_similarity instance.

        Args:
            test_cases (list): List of test cases to evaluate.
            number_of_prompts (int): Number of prompts to generate and/or test.
            model_test (str): The language model used for testing.
            model_test_temperature (float): The temperature parameter for the testing model.
            model_test_max_tokens (int): The maximum number of tokens allowed for the testing model.
            model_generation (str): The language model used for generating prompts.
            model_generation_temperature (float): The temperature parameter for the generation model.
            prompts (list): List of prompts to evaluate.
            best_prompts (int): Number of best prompts to consider.

        Note:
            The 'system_gen_system_prompt' attribute is predefined within the class constructor.
        """

        self.test_cases = test_cases
        self.model_test = model_test
        self.model_test_temperature = model_test_temperature
        self.model_test_max_tokens = model_test_max_tokens
        self.system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Specify in the prompts you generate that the response must be text.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
        self.prompts = prompts
        self.model_embedding = model_embedding
        self.best_prompts = best_prompts
        self.timeout = timeout
        self.n_retries = n_retries
    
    def process_prompt(self, prompt, test_case, model, model_max_tokens, model_temperature):
        messages = [
            {"role": "system", "content": prompt},
            {"role": "user", "content": f"{test_case['input']}"}
        ]
        response = openai_call.create_chat_completion(model, messages, model_max_tokens, model_temperature, 1, timeout=self.timeout, n_retries=self.n_retries)
        partial_tokens_input = response.usage.prompt_tokens
        partial_tokens_output = response.usage.completion_tokens
        result_content = response.choices[0].message.content
        
        return partial_tokens_input, partial_tokens_output, result_content, test_case['output']
    
    def calculate_embedding(self, test_case):
        response_embedding = openai_call.create_embedding(self.model_embedding, test_case['output'], self.timeout, self.n_retries)
        test_case['embedding'] = response_embedding.data[0].embedding
        tokens_embedding = response_embedding.usage.total_tokens
        return tokens_embedding
    
    def test_candidate_prompts(self):

        """
        Test a list of candidate prompts with test cases and evaluate their performance.

        Returns:
            tuple: A tuple containing the following elements:
                - List of results and statistics.
                - List of best-performing prompts.
                - Cost of generating and testing prompts.
                - Tokens used for input.
                - Tokens used for output.
                - Tokens used for embeddings.
        """

        cost = 0
        tokens_input = 0
        tokens_output = 0
        tokens_embedding = 0
        prompt_results = {prompt: {'score': 0} for prompt in self.prompts}
        results = [{"method": "Semantic Similarity"}]

        if 'embedding' not in self.test_cases[0]:
            threads = []

            results_embedding = []

            for test_case in self.test_cases:
                thread = threading.Thread(target=lambda t=test_case: results_embedding.append(self.calculate_embedding(t)))
                threads.append(thread)
                thread.start()

            for thread in threads:
                thread.join()

            tokens_embedding = sum(results_embedding)


        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []

            for prompt in self.prompts:
                prompt_and_results = [{"prompt": prompt}]
                for test_case in self.test_cases:
                    future = executor.submit(
                        self.process_prompt,
                        prompt,
                        test_case,
                        self.model_test,
                        self.model_test_max_tokens,
                        self.model_test_temperature
                    )
                    futures.append(future)
                    partial_tokens_input, partial_tokens_output, result_content, ideal_output = future.result()
                    tokens_input += partial_tokens_input
                    tokens_output += partial_tokens_output
                    embedding_response = openai_call.create_embedding(self.model_embedding, result_content, self.timeout, self.n_retries)
                    embedding_test_case = test_case['embedding']
                    similarity_score = cosine_similarity(np.array(embedding_response.data[0].embedding).reshape(1, -1) , np.array(embedding_test_case).reshape(1, -1))[0, 0]
                    
                    prompt_results[prompt]['score'] = prompt_results[prompt]['score'] + similarity_score
                    tokens_embedding = tokens_embedding + embedding_response.usage.total_tokens

                    prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": similarity_score})
                
                results.append(prompt_and_results)
                prompt_and_results = []

        cost_input = input.cost(tokens_input, self.model_test)
        cost_output = output.cost(tokens_output, self.model_test)
        cost_embeddings = embeddings.cost(tokens_embedding, self.model_embedding)
        cost = cost_input + cost_output + cost_embeddings

        # Sort the prompts by score.
        data_list = []
        for i, prompt in enumerate(self.prompts):
            score = prompt_results[prompt]['score']/len(self.test_cases)
            data_list.append({"prompt": prompt, "rating": score})

        sorted_data = sorted(data_list, key=lambda x: x['rating'], reverse=True)
        best_prompts = sorted_data[:self.best_prompts]
        sorted_data.append(results)
        return sorted_data, best_prompts, cost, tokens_input, tokens_output, tokens_embedding
    
    def evaluate_optimal_prompt(self):

        """
        Evaluate the optimal prompt by testing candidate prompts and selecting the best ones.

        Returns:
            tuple: A tuple containing the result data, best prompts, cost, input tokens used, output tokens and embedding tokens used.
        """
        
        return self.test_candidate_prompts()


from types import SimpleNamespace

constants = SimpleNamespace(
    OPENAI_API_BASE="OPENAI_API_BASE",
    OPENAI_API_KEY="OPENAI_API_KEY",
    OPENAI_API_TYPE="OPENAI_API_TYPE",
    OPENAI_API_VERSION="OPENAI_API_VERSION"
)


def cost (tokens, model):
    """
    Calculate the cost (in USD) of generating a given number of tokens using a specified language model.

    Args:
        tokens (int): The number of tokens to calculate the cost for.
        model (str): The language model to use for cost calculation (e.g., 'text-embedding-ada-002'.

    Returns:
        float: The cost in USD for generating the specified number of tokens with the chosen model.
    """
    if tokens == 0:
        cost = 0

    else:
        
        if model == 'text-embedding-ada-002':
            cost = (0.0001/1000)*tokens

        if model == 'text-similarity-ada-001':
            cost = (0.004/1000)*tokens

        if model == 'text-search-ada-query-001':
            cost = (0.004/1000)*tokens

        if model == 'code-search-ada-code-001':
            cost = (0.004/1000)*tokens

        if model == 'code-search-ada-text-001':
            cost = (0.004/1000)*tokens

        if model == 'text-similarity-babbage-001' or model == 'text-search-babbage-doc-001' or model == 'text-search-babbage-query-001' or model == 'code-search-babbage-code-001' or model == 'code-search-babbage-text-001':
            cost = (0.005/1000)*tokens

        if model == 'text-similarity-curie-001' or model == 'text-search-curie-doc-001' or model == 'text-search-curie-query-001':
            cost = (0.02/1000)*tokens

        if model == 'text-similarity-davinci-001' or model == 'text-search-davinci-doc-001' or model == 'text-search-davinci-query-001':
            cost = (0.2/1000)*tokens

    return cost


from ..openai_calls import openai_call
from ..cost import input, output
import concurrent.futures
from typing import List, Dict
import math

class LogProbs:
    def __init__(self, test_cases: List[Dict], prompts: List[str], model_test: str="gpt-3.5-turbo-instruct", model_test_temperature: float=0.6, model_test_max_tokens: int=4, best_prompts: int=2, timeout: int=10, n_retries: int=5):

        """
        Initialize a LogProbs instance.

        Args:
            test_cases (list): List of test cases to evaluate.
            number_of_prompts (int): Number of prompts to generate and/or test.
            model_test (str): The language model used for testing.
            model_test_temperature (float): The temperature parameter for the testing model.
            model_test_max_tokens (int): The maximum number of tokens allowed for the testing model.
            model_generation (str): The language model used for generating prompts.
            model_generation_temperature (float): The temperature parameter for the generation model.
            prompts (list): List of prompts to evaluate.
            best_prompts (int): Number of best prompts to consider.

        Note:
            The 'system_gen_system_prompt' attribute is predefined within the class constructor.
        """

        self.test_cases = test_cases
        self.model_test = model_test
        self.model_test_temperature = model_test_temperature
        self.model_test_max_tokens = model_test_max_tokens
        self.system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Specify in your prompts that the response has to be with the token with the highest logprobs and the response should noy have \n at the beginning.

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
        self.prompts = prompts
        self.best_prompts = best_prompts
        self.timeout = timeout
        self.n_retries = n_retries

    def process_prompt(self, prompt, test_case, model, model_max_tokens, model_temperature):
        messages = prompt + ' ' + test_case['input']
        response = openai_call.create_completion(model, messages, model_max_tokens, model_temperature, 1, timeout=self.timeout, n_retries=self.n_retries)
        partial_tokens_input = response.usage.prompt_tokens
        partial_tokens_output = response.usage.completion_tokens
        top_logprobs = response.choices[0].logprobs.top_logprobs
        
        return partial_tokens_input, partial_tokens_output, top_logprobs

    def test_candidate_prompts(self):

        """
        Test a list of candidate prompts with test cases and evaluate their performance.

        Returns:
            tuple: A tuple containing the following elements:
                - List of results and statistics.
                - List of best-performing prompts.
                - Cost of generating and testing prompts.
                - Tokens used for input.
                - Tokens used for output.
        """

        cost = 0
        tokens_input = 0
        tokens_output = 0
        prompt_results = {prompt: {'total': 0} for prompt in self.prompts}
        results = [{"method": "LogProbs"}]

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []

            for prompt in self.prompts:
                prompt_and_results = [{"prompt": prompt}]
                for test_case in self.test_cases:
                    future = executor.submit(
                        self.process_prompt,
                        prompt,
                        test_case,
                        self.model_test,
                        self.model_test_max_tokens,
                        self.model_test_temperature
                    )
                    futures.append(future)
                    partial_tokens_input, partial_tokens_output, logprobs = future.result()
                    tokens_input += partial_tokens_input
                    if type(partial_tokens_output) == int:
                        tokens_output += partial_tokens_output

                    def generate_combinations(dictionaries, index=0, current_combination="", current_percentage=1, memo={}):
                        if index == len(dictionaries):
                            return [{current_combination: current_percentage}]

                        if (index, current_combination) in memo:
                            return memo[(index, current_combination)]

                        results = []
                        for key, value in dictionaries[index].items():
                            new_combination = current_combination + key
                            new_percentage = current_percentage * math.exp(value)
                            results += generate_combinations(dictionaries, index + 1, new_combination, new_percentage, memo)

                        memo[(index, current_combination)] = results
                        return results
                    
                    def clean_keys(array):
                        for dictionary in array:
                            original_keys = list(dictionary.keys())
                            for key in original_keys:
                                new_key = key.replace('\n', '').replace('', '').replace('"', '').replace("'", '').replace(',', '').replace('<|endoftext|>', '').replace(':', '').lstrip()
                                dictionary[new_key] = dictionary.pop(key)
                        return array
                    
                    def sum_repeated_values(array):
                        result_dictionary = {}

                        for dictionary in array:
                            for key, value in dictionary.items():
                                key_lower = key.lower()
                                if key_lower in result_dictionary:
                                    result_dictionary[key_lower] += value
                                else:
                                    result_dictionary[key_lower] = value

                        result = [{'{}'.format(key): value} for key, value in result_dictionary.items()]
                        return result
                    
                    all_combinations = generate_combinations(logprobs)

                    clean_combinations = clean_keys(all_combinations)

                    combinations = sum_repeated_values(clean_combinations)

                    for diccionario in combinations:
                        for key, value in diccionario.items():
                            if test_case['output'].lower() == key:
                                probability = value
                                break

                    
                    if not(any(test_case['output'].lower() in diccionario for diccionario in combinations)):
                        probability = 0

                    prompt_results[prompt]['total'] = prompt_results[prompt]['total'] + probability
                    prompt_and_results.append({"test": test_case['input'], "probability": probability, "answer": test_case["output"]})
                
                results.append(prompt_and_results)
                prompt_and_results = []

        cost_input = input.cost(tokens_input, self.model_test)
        cost_output = output.cost(tokens_output, self.model_test)
        cost = cost + cost_input + cost_output

        
        # Sort the prompts by score.
        data_list = []
        for i, prompt in enumerate(self.prompts):
            score = prompt_results[prompt]['total']/len(self.test_cases)
            data_list.append({"prompt": prompt, "rating": score})

        sorted_data = sorted(data_list, key=lambda x: x['rating'], reverse=True)
        best_prompts = sorted_data[:self.best_prompts]
        sorted_data.append(results)
        return sorted_data, best_prompts, cost, tokens_input, tokens_output
    
    def evaluate_optimal_prompt(self):

        """
        Evaluate the optimal prompt by testing candidate prompts and selecting the best ones.

        Returns:
            tuple: A tuple containing the result data, best prompts, cost, input tokens used, and output tokens used.
        """
        
        return self.test_candidate_prompts()


from ..openai_calls import openai_call
from ..cost import input, output
import ast
import concurrent.futures
import builtins
import js2py
import re
import os
from typing import List, Dict

class codeGeneration:
    def __init__(self, test_cases: List[Dict], prompts: List[str], model_test: str='gpt-3.5-turbo', model_test_temperature: float=0.6, model_test_max_tokens: int=1000, best_prompts: int=2, timeout: int=10, n_retries: int=5):

        """
        Initialize a Code Generation instance.

        Args:
            test_cases (list): List of test cases to evaluate.
            number_of_prompts (int): Number of prompts to generate and/or test.
            model_test (str): The language model used for testing.
            model_test_temperature (float): The temperature parameter for the testing model.
            model_test_max_tokens (int): The maximum number of tokens allowed for the testing model.
            model_generation (str): The language model used for generating prompts.
            model_generation_temperature (float): The temperature parameter for the generation model.
            prompts (list): List of prompts to evaluate.
            best_prompts (int): Number of best prompts to consider.

        Note:
            The 'system_gen_system_prompt' attribute is predefined within the class constructor.
        """

        self.test_cases = test_cases
        self.model_test = model_test
        self.model_test_temperature = model_test_temperature
        self.model_test_max_tokens = model_test_max_tokens
        self.system_gen_system_prompt = """Your job is to generate system prompts for GPT, given a description of the use-case and some test cases.

        In your generated prompt, you should describe how the AI should behave in plain English. Include what it will see, and what it's allowed to output. Be creative in with prompts to get the best possible results. The AI knows it's an AI -- you don't need to tell it this.

        You will be graded based on the performance of your prompt... but don't cheat! You cannot include specifics about the test cases in your prompt. Any prompts with examples will be disqualified.

        Specify in your prompts that you are going to generate that your task is to be a code generator, so the result to the prompt you are going to generate can only be a script in some programming language and the arguments you are going to use in your created function.

        These are examples: "def count(l):\n return len(l)", "def suma_lista(lista):\n total = 0\n for elemento in lista:\n\n total += elemento\n return total", "def factorial(n):\n if n == 0:\n\n return 1\n else:\n\n return n * factorial(n - 1)", "def es_par(numero):\n return numero % 2 == 0", "def maximo_lista(lista):\n if len(lista) == 0:\n\n return None\n else:\n\n return max(lista)", That is the only valid output format for the prompt.", Note you don't need to use print() or console.log().

        Most importantly, output NOTHING but the prompt. Do not include anything else in your message."""
        self.prompts = prompts
        self.best_prompts = best_prompts
        self.timeout = timeout
        self.n_retries = n_retries

    def process_prompt(self, prompt, test_case, model, model_max_tokens, model_temperature):
        messages = [
            {"role": "system", "content": prompt},
            {"role": "user", "content": f"{test_case['input']}"}
        ]
        response = openai_call.create_chat_completion(model, messages, model_max_tokens, model_temperature, 1, timeout=self.timeout, n_retries=self.n_retries)
        partial_tokens_input = response.usage.prompt_tokens
        partial_tokens_output = response.usage.completion_tokens
        result_content = response.choices[0].message.content
        
        return partial_tokens_input, partial_tokens_output, result_content, test_case['output']

    def test_candidate_prompts(self):

        """
        Test a list of candidate prompts with test cases and evaluate their performance.

        Returns:
            tuple: A tuple containing the following elements:
                - List of results and statistics.
                - List of best-performing prompts.
                - Cost of generating and testing prompts.
                - Tokens used for input.
                - Tokens used for output.
        """

        cost = 0
        tokens_input = 0
        tokens_output = 0
        prompt_results = {prompt: {'correct': 0, 'total': 0} for prompt in self.prompts}
        results = [{"method": "Code Generation"}]

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
        
            for prompt in self.prompts:
                prompt_and_results = [{"prompt": prompt}]
                for test_case in self.test_cases:
                    
                    future = executor.submit(
                        self.process_prompt,
                        prompt,
                        test_case,
                        self.model_test,
                        self.model_test_max_tokens,
                        self.model_test_temperature
                    )
                    futures.append(future)
                    partial_tokens_input, partial_tokens_output, result_content, ideal_output = future.result()
                    tokens_input += partial_tokens_input
                    tokens_output += partial_tokens_output  
                    
                    if 'python' in test_case['input'].lower():
                        try:
                        
                            
                            with open("functions.py", "w") as file:
                                parsed_code = ast.parse(result_content)
                                has_function = any(isinstance(stmt, ast.FunctionDef) for stmt in parsed_code.body)
                                if has_function:
                                    
                                    file.write(ast.unparse(parsed_code))
                            
                            
                            with open("functions.py", "r") as file:
                                function_code = file.read()
                        
                        
                            exec(function_code, builtins.__dict__)

                        
                            name_function = None
                            for stmt in reversed(parsed_code.body):
                                if isinstance(stmt, ast.FunctionDef):
                                    name_function = stmt.name
                                    break
                            if name_function is not None:

                                
                                executable_function = builtins.__dict__[name_function]

                            
                                result = executable_function(*ast.literal_eval(test_case['arguments'])) 
                            
                            
                            if result == ideal_output:
                                prompt_results[prompt]['correct'] += 1
                                prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": True})
                                prompt_results[prompt]['total'] += 1
                            if result != ideal_output:
                                prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": "Final result incorrect."})
                                prompt_results[prompt]['total'] += 1

                        except Exception as e:
    
                            prompt_results[prompt]['total'] += 1
                            prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": "Not Python script."})
                    
                    if 'javascript' in test_case['input'].lower():
                        
                        try:
                            file = "functions.js"
                            with open("functions.js", "w") as file:
                                file.write(result_content)

                            with open("functions.js", "r") as f:
                                js_code = f.read()
                            
                            context = js2py.EvalJs()

                            context.execute(js_code)

                            functions = re.findall(r"function\s+(\w+)\s*\(", js_code)

                            for function_name in functions:
                                
                                js_function = context.eval(function_name)
                                result = js_function(*ast.literal_eval(test_case['arguments']))
                            
                                def convert_to_native(js_object, target_type):
                                
                                    conversion_functions = {
                                        int: lambda x: int(x),
                                        float: lambda x: float(x),
                                        str: lambda x: str(x),
                                        bool: lambda x: bool(x),
                                        list: lambda x: list(x),
                                        dict: lambda x: dict(x),
                                        tuple: lambda x: tuple(x),
                            
                                    }

                                
                                    if target_type in conversion_functions:
                                
                                        conversion_function = conversion_functions[target_type]
                                        result = conversion_function(js_object)
                                        return result
                                    
                                final_result = convert_to_native(result, type(ideal_output))

                            if final_result == ideal_output:
                                prompt_results[prompt]['correct'] += 1
                                prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": True})
                                prompt_results[prompt]['total'] += 1
                            if final_result != ideal_output:
                                prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": "Final result incorrect."})
                                prompt_results[prompt]['total'] += 1

                        except Exception as e:

                            prompt_results[prompt]['total'] += 1
                            prompt_and_results.append({"test": test_case['input'], "answer": result_content, "ideal": ideal_output, "result": "Not Javascript script."})

                results.append(prompt_and_results)
                prompt_and_results = []

        cost_input = input.cost(tokens_input, self.model_test)
        cost_output = output.cost(tokens_output, self.model_test)
        cost = cost + cost_input + cost_output

        # Calculate and print the percentage of correct answers and average time for each model
        best_prompt = self.prompts[0]
        best_percentage = 0
        data_list = []
        for i, prompt in enumerate(self.prompts):
            correct = prompt_results[prompt]['correct']
            total = prompt_results[prompt]['total']
            percentage = (correct / total) * 100
            data_list.append({"prompt": prompt, "rating": percentage})
            print(f"Prompt {i+1} got {percentage:.2f}% correct.")
            if percentage >= best_percentage:
                best_percentage = percentage
                best_prompt = prompt
        sorted_data = sorted(data_list, key=lambda x: x['rating'], reverse=True)
        best_prompts = sorted_data[:self.best_prompts]
        print(f"The best prompt was '{best_prompt}' with a correctness of {best_percentage:.2f}%.")
        sorted_data.append(results)
        return sorted_data, best_prompts, cost, tokens_input, tokens_output
    
    def evaluate_optimal_prompt(self):

        """
        Evaluate the optimal prompt by testing candidate prompts and selecting the best ones.

        Returns:
            tuple: A tuple containing the result data, best prompts, cost, input tokens used, and output tokens used.
        """

        result = self.test_candidate_prompts()
        file_to_delete_python = "functions.py"
        if os.path.isfile(file_to_delete_python):
            os.unlink(file_to_delete_python)
        file_to_delete_javascript = "functions.js"
        if os.path.isfile(file_to_delete_javascript):
            os.unlink(file_to_delete_javascript)

        return result


def cost (tokens, model):
    """
    Calculate the cost (in USD) of generating a given number of tokens using a specified language model.

    Args:
        tokens (int): The number of tokens to calculate the cost for.
        model (str): The language model to use for cost calculation (e.g., 'gpt-3.5-turbo', 'gpt-4').

    Returns:
        float: The cost in USD for generating the specified number of tokens with the chosen model.
    """
    if tokens == 0:
        cost = 0
    if model == 'gpt-3.5-turbo':
        cost = (0.002 /1000) * tokens
    if model == 'gpt-3.5-turbo-16K':
        cost = (0.004/1000) * tokens
    if model == 'gpt-4':
        cost = (0.06/1000) * tokens
    if model == 'gpt-4-32K':
        cost = (0.12/1000) * tokens
    if model == 'gpt-3.5-turbo-instruct':
        cost = (0.002 /1000) * tokens
    if model == 'gpt-4-turbo':
        cost = (0.03/1000) * tokens
    return cost


from setuptools import setup

with open('promptwizard/doc/doc.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='promptwizard',
    version='0.0.15',
    packages=['promptwizard', 'promptwizard.evals', 'promptwizard.cost', 'promptwizard.approximate_cost', 'promptwizard.validation_yaml', 'promptwizard.prompt_generation', 'promptwizard.openai_calls'],
    python_requires='>=3.8',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'openai==1.2.2',
        'PyYAML>=5.4',
        'matplotlib',
        'numpy',
        'marshmallow>=3.20.1',
        'tiktoken>=0.4.0',
        'tenacity>=8.2.3',
        'zipp',
        'six',
        'aiohttp',
        'urllib3',
        'certifi',
        'tqdm',
        'python-dotenv',
        'js2py',
        'scikit-learn',
        'cohere==4.34',
        'httpx'
    ],
    entry_points={
        'console_scripts': [
            'promptwizard=promptwizard.main:main',
        ],
    },
    package_data={'': ['doc/doc.md']},
    include_package_data=True,
    description='Prompt Wizard is a package for evaluating custom prompts using various evaluation methods. It allows you to provide your own prompts or generate them automatically and then obtain the results in a JSON file. You can also use the library for your Python projects.',
    license='MIT',
)


from promptwizard.cost import input, output
from ..openai_calls import openai_call
from typing import List, Dict

def generate_candidate_prompts(system_gen_system_prompt: str, test_cases: List[Dict], description: str, model_generation: str='gpt-4-turbo', model_generation_temperature: float=1.2, model_generation_max_tokens: int=300, number_of_prompts: int=4, prompt_features: str=None, timeout: int=10, n_retries: int=5):
    try:    
        # Check if additional prompt features are provided
        if prompt_features is not None:
            # Construct messages for generating prompts with prompt features
            model=model_generation
            messages = [
                {"role": "system", "content": system_gen_system_prompt + ' ' + prompt_features},
                {"role": "user", "content": f"Here are some test cases:`{test_cases}`\n\nHere is the description of the use-case: `{description.strip()}`\n\nRespond with your prompt, and nothing else. Be creative. Amplify your vocabulary as much as you can. All your prompts must be different from each other."}
                ]
            temperature=model_generation_temperature
            n=number_of_prompts
            max_tokens=model_generation_max_tokens
            outputs = openai_call.create_chat_completion(model, messages, max_tokens, temperature, n, timeout=timeout, n_retries=n_retries)
        else:
            # Construct messages for generating prompts without prompt features
            model=model_generation
            messages=[
                {"role": "system", "content": system_gen_system_prompt},
                {"role": "user", "content": f"Here are some test cases:`{test_cases}`\n\nHere is the description of the use-case: `{description.strip()}`\n\nRespond with your prompt, and nothing else. Be creative. Amplify your vocabulary as much as you can. All your prompts must be different from each other."}
                ]
            temperature=model_generation_temperature
            n=number_of_prompts
            max_tokens=model_generation_max_tokens
            # Generate prompts using OpenAI's ChatCompletion API
            outputs = openai_call.create_chat_completion(model, messages, max_tokens, temperature, n, timeout=timeout, n_retries=n_retries)

        # Extract tokens input and tokens output usage
        tokens_input = outputs.usage.prompt_tokens
        tokens_output = outputs.usage.completion_tokens

        # Calculate the cost based on token usage
        cost_input = input.cost(tokens_input, model_generation)
        cost_output = output.cost(tokens_output, model_generation)
        cost = cost_input + cost_output

        # Extract the generated prompts
        prompts = [choice.message.content for choice in outputs.choices]
        return prompts, cost, tokens_input, tokens_output
    
    except Exception as e:
        print(f"An exception occurred: {e}")
        raise


import openai
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_not_exception_type,
)
from typing import List
from openai import OpenAI
import httpx





def create_chat_completion(model: str, messages :List[dict], max_tokens: int, temperature: float, number_of_prompts: int, logit_bias: dict=None, functions: dict=None, function_call: str=None, timeout: int=10, n_retries: int=5):

    if model == 'gpt-4-turbo':
            model = 'gpt-4-1106-preview'

    client = OpenAI(timeout=httpx.Timeout(timeout, read=5.0, write=10.0, connect=3.0))
            
    @retry(wait=wait_random_exponential(min=1, max=timeout), stop=stop_after_attempt(n_retries), retry=retry_if_not_exception_type(openai.BadRequestError))
    def create_chat_completion_retry():
            
        try:
            if (logit_bias==None and functions==None):

                respond = client.chat.completions.create(
                    model=model,
                    messages=messages,
                    max_tokens=max_tokens,
                    n = number_of_prompts,
                    temperature=temperature,
                )

            elif functions!=None:
                    
                respond = client.chat.completions.create(
                    model=model,
                    messages=messages,
                    max_tokens=max_tokens,
                    n = number_of_prompts,
                    temperature=temperature,
                    functions=functions,
                    function_call=function_call,
                )

            else:

                respond = client.chat.completions.create(
                    model=model,
                    messages=messages,
                    max_tokens=max_tokens,
                    n = number_of_prompts,
                    temperature=temperature,
                    logit_bias=logit_bias,
                )
        
        except openai.APITimeoutError as e:
            print(f"API Timeout Error: {e}")
        except openai.APIError as err:
            print(f"API Error: {err}")
        except openai.OpenAIError as e:
            print(f"Error in request: {e}")
            raise
        except Exception as err:
            print(f"An unexpected error occurred: {e}")
                
        return respond
    return create_chat_completion_retry()

def create_embedding(model: str, input: str, timeout: int=10, n_retries: int=5):

    client = OpenAI(timeout=httpx.Timeout(timeout, read=5.0, write=10.0, connect=3.0))

    @retry(wait=wait_random_exponential(min=1, max=timeout), stop=stop_after_attempt(n_retries), retry=retry_if_not_exception_type(openai.BadRequestError))
    def create_embedding_retries():

        try:
            embedding = client.embeddings.create(
                model=model,
                input=input
            )

        except openai.OpenAIError as e:
            print(f"Error in request: {e}")
            raise

        return embedding
    return create_embedding_retries()

def create_completion(model: str, messages: str, max_tokens: int, temperature: float, number_of_prompts: int, timeout: int=10, n_retries: int=5):

    client = OpenAI(timeout=httpx.Timeout(timeout, read=5.0, write=10.0, connect=3.0))

    @retry(wait=wait_random_exponential(min=1, max=timeout), stop=stop_after_attempt(n_retries), retry=retry_if_not_exception_type(openai.BadRequestError))
    def create_completion_retry():
    
        try:

            respond = client.completions.create(
                model=model,
                prompt=messages,
                max_tokens=max_tokens,
                n = number_of_prompts,
                temperature=temperature,
                logprobs=5,
            )
        
        except openai.OpenAIError as e:
            print(f"Error in request: {e}")
            raise
                
        return respond
    return create_completion_retry()


from . import generation
from promptwizard.evals import elovalue, classification, equals, includes, function_calling, code_generation, json_validation, semantic_similarity, logprobs
from typing import List, Dict

def iterations(test_cases: List[Dict], method: str, old_prompts_and_rating: List[str], new_number_of_prompts: int, functions: Dict=None, model_test: str='gpt-3.5-turbo', model_test_temperature: float=1.2, model_test_max_tokens: int=1000, model_generation: str='gpt-4-turbo', model_generation_temperature: float=0.6, model_generation_max_tokens: int=500, function_call: str='auto', description: str=None, best_prompts: int=2, model_embeddings: str='text-embedding-ada-002', timeout: int=10, n_retries: int=5):

    # Initialize cost and token variables
    cost = 0
    tokens_input_gen = 0
    tokens_output_gen = 0
    tokens_input_test = 0
    tokens_output_test = 0

    # Determine the class_method based on the specified method
    if method == 'Elo':
        class_method = elovalue.Elo
    if method == 'Classification':
        class_method = classification.Classification
    if method == 'Equals':
        class_method = equals.Equals
    if method == 'Includes':
        class_method = includes.Includes
    if method == 'Function Calling':
         class_method = function_calling.functionCalling
    if method == 'Code Generation':
        class_method = code_generation.codeGeneration
    if method == 'JSON Validation':
        class_method = json_validation.jsonValidation

    if method == 'Semantic Similarity':
         class_method = semantic_similarity.semanticSimilarity
    if method == 'LogProbs':
         class_method = logprobs.LogProbs

    # Generate new prompts if the method is not 'Elo'
    if method != 'Elo' and method != 'Semantic Similarity':
            new_prompts = []
            old_prompts = []

            # Extract old prompts from old_prompts_and_rating
            for item in old_prompts_and_rating:
                prompt_content = item["prompt"]
                old_prompts.append(prompt_content)

            candidate_prompts = []
            # Generate candidate prompts for the next iteration
            candidates_iteration = generation.generate_candidate_prompts("Your job is to generate a similar prompt to prompts you are going to receive. Generate a new one by modifying words or phrases but in such a way that the meaning of the prompt is preserved. What you return has to be a reformulation of what you received and nothing more, no explanation is necessary. Don't return phrases like 'Here are some examples:', only provide a prompt, nothing else.", old_prompts, 'Generate a new prompt from the ones written here.', model_generation, model_generation_temperature, model_generation_max_tokens, number_of_prompts=new_number_of_prompts, prompt_features=None, timeout=10, n_retries=5)
            candidates = candidates_iteration[0]
            cost = cost + candidates_iteration[1]
            tokens_input_gen = tokens_input_gen + candidates_iteration[2]
            tokens_output_gen = tokens_output_gen + candidates_iteration[3]
            candidate_prompts.extend(candidates)

            new_prompts.extend(candidate_prompts)

            # Determine which evaluation method to use and evaluate optimal prompts
            if method == 'Function Calling':
                evaluable_object = class_method(test_cases, new_prompts, functions, model_test, model_test_temperature, model_test_max_tokens, best_prompts, function_call, timeout, n_retries)
            if method != 'Function Calling':
                evaluable_object = class_method(test_cases, new_prompts, model_test, model_test_temperature, model_test_max_tokens, best_prompts, timeout, n_retries)
            results = evaluable_object.evaluate_optimal_prompt()
            cost = cost + results[2]
            tokens_input_test = tokens_input_test + results[3]
            tokens_output_test = tokens_output_test + results[4]

    if method == 'Semantic Similarity':
            new_prompts = []
            old_prompts = []

            # Extract old prompts from old_prompts_and_rating
            for item in old_prompts_and_rating:
                prompt_content = item["prompt"]
                old_prompts.append(prompt_content)

            candidate_prompts = []
            # Generate candidate prompts for the next iteration
            candidates_iteration = generation.generate_candidate_prompts("Your job is to generate a similar prompt to prompts you are going to receive. Generate a new one by modifying words or phrases but in such a way that the meaning of the prompt is preserved. What you return has to be a reformulation of what you received and nothing more, no explanation is necessary. Don't return phrases like 'Here are some examples:', only provide a prompt, nothing else.", old_prompts, 'Generate a new prompt from the ones written here.', model_generation, model_generation_temperature, model_generation_max_tokens, number_of_prompts=new_number_of_prompts, prompt_features=None, timeout=10, n_retries=5)
            candidates = candidates_iteration[0]
            cost = cost + candidates_iteration[1]
            tokens_input_gen = tokens_input_gen + candidates_iteration[2]
            tokens_output_gen = tokens_output_gen + candidates_iteration[3]
            candidate_prompts.extend(candidates)

            new_prompts.extend(candidate_prompts)

            evaluable_object = class_method(test_cases, new_prompts, model_test, model_test_temperature, model_test_max_tokens, model_embeddings, timeout, n_retries)
            results = evaluable_object.evaluate_optimal_prompt()
            cost = cost + results[2]
            tokens_input_test = tokens_input_test + results[3]
            tokens_output_test = tokens_output_test + results[4]
            tokens_embeddings = results[5]

    # If the method is 'Elo', generate new prompts differently
    if method=='Elo':
            new_prompts = []
            candidate_prompts = []
            # Generate candidate prompts for the next iteration
            candidates_iteration = generation.generate_candidate_prompts("Your job is to generate a similar prompt to prompts you are going to receive. Generate a new one by modifying words or phrases but in such a way that the meaning of the prompt is preserved. What you return has to be a reformulation of what you received and nothing more, no explanation is necessary. Don't return phrases like 'Here are some examples:', just say the prompt", old_prompts_and_rating, 'Generate a new prompt from the ones written here.', model_generation, model_generation_temperature, model_generation_max_tokens, number_of_prompts=new_number_of_prompts, prompt_features=None, timeout=10, n_retries=5)
            candidates = candidates_iteration[0]
            cost = cost + candidates_iteration[1]
            candidate_prompts.extend(candidates)

            # Include old prompts as candidates
            for prompt in old_prompts_and_rating:
                 candidate_prompts.append(prompt)

            # Evaluate optimal prompts using the 'Elo' method
            evaluable_object = class_method(test_cases, candidate_prompts, description, model_test, model_test_temperature, model_test_max_tokens, timeout, n_retries)
            results = evaluable_object.evaluate_optimal_prompt()
            cost = cost + results[2]
            tokens_input_test = tokens_input_test + results[3]
            tokens_output_test = tokens_output_test + results[4]

    if method == 'Semantic Similarity':
         return results, cost, tokens_input_gen, tokens_input_test, tokens_output_gen, tokens_output_test, tokens_embeddings
    return results, cost, tokens_input_gen, tokens_input_test, tokens_output_gen, tokens_output_test

def iterate(test_cases, method, prompts, iteration, best_percentage=100, best_prompts=2, functions: Dict=None, model_test: str='gpt-3.5-turbo', model_test_temperature: float=1.2, model_test_max_tokens: int=1000, model_iteration: str='gpt-4-turbo', model_iteration_temperature: float=0.6, model_iteration_max_tokens: int=500, function_call: str='auto', description: str=None, model_embeddings: str='text-embedding-ada-002', timeout: int=10, n_retries: int=5):
    old_prompts = prompts[:best_prompts]
    number_of_prompts = len(old_prompts)
    final_results = []
    number_of_iteration = 0
    if method != 'Elo' and method != 'Semantic Similarity':
        tokens_input_gen_iter = 0
        tokens_output_gen_iter = 0
        tokens_input_test_iter = 0
        tokens_output_test_iter = 0
        all_have_rating_percentage = True
        # Iterate through the list of elements and check the 'rating' value.
        for element in old_prompts:
            if element["rating"] < best_percentage:
                all_have_rating_percentage = False
                break  # If an element with a different rating is found, stop the iteration.
        while (iteration > 0 and (not all_have_rating_percentage)):
            print(f"Iteration number {number_of_iteration}")
            combine_prompts = []
            if method == 'Function Calling':
                new_results_prompts_cost = iterations(test_cases, method, old_prompts, number_of_prompts - best_prompts, functions, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, function_call, None, best_prompts, timeout, n_retries)
                final_results.append(new_results_prompts_cost)

            if method != 'Function Calling':
                new_results_prompts_cost = iteration.iterations(test_cases, method, old_prompts, number_of_prompts - best_prompts, None, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, None, None, best_prompts, timeout, n_retries)
                final_results.append(new_results_prompts_cost)
            new_results = new_results_prompts_cost[0]
            cost = cost + new_results_prompts_cost[1]
            tokens_input_gen_iter = tokens_input_gen_iter + new_results_prompts_cost[2]
            tokens_output_gen_iter = tokens_output_gen_iter + new_results_prompts_cost[4]
            tokens_input_test_iter = tokens_input_test_iter + new_results_prompts_cost[3]
            tokens_output_test_iter = tokens_output_test_iter + new_results_prompts_cost[5]
            iteration = iteration - 1
            number_of_iteration = number_of_iteration + 1
            combine_prompts.append(old_prompts)
            combine_prompts.append(new_results[1])
            combined_data = [item for sublist in combine_prompts for item in sublist]
            sorted_data = sorted(combined_data, key=lambda x: x['rating'], reverse=True)
            old_prompts = sorted_data[:best_prompts]
            all_have_rating_percentage = True
            for element in old_prompts:
                if element["rating"] < best_percentage:
                    all_have_rating_percentage = False
                    break  # If an element with a different rating is found, stop the iteration.

    if method == 'Semantic Similarity':
        final_results = []
        tokens_input_gen_iter = 0
        tokens_output_gen_iter = 0
        tokens_input_test_iter = 0
        tokens_output_test_iter = 0
        tokens_embeddings_iter = 0
        while (iteration > 0):
            print(f"Iteration number {number_of_iteration}")
            combine_prompts = []

            new_results_prompts_cost = iterations(test_cases, method, old_prompts, number_of_prompts - best_prompts, None, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, None, None, best_prompts, model_embeddings, timeout, n_retries)
            final_results.append(new_results_prompts_cost)
            new_results = new_results_prompts_cost[0]
            cost = cost + new_results_prompts_cost[1]
            tokens_input_gen_iter = tokens_input_gen_iter + new_results_prompts_cost[2]
            tokens_output_gen_iter = tokens_output_gen_iter + new_results_prompts_cost[4]
            tokens_input_test_iter = tokens_input_test_iter + new_results_prompts_cost[3]
            tokens_output_test_iter = tokens_output_test_iter + new_results_prompts_cost[5]
            tokens_embeddings_iter = tokens_embeddings_iter + new_results_prompts_cost[6]
            iteration = iteration - 1
            number_of_iteration = number_of_iteration + 1
            combine_prompts.append(old_prompts)
            combine_prompts.append(new_results[1])
            combined_data = [item for sublist in combine_prompts for item in sublist]
            sorted_data = sorted(combined_data, key=lambda x: x['rating'], reverse=True)
            old_prompts = sorted_data[:best_prompts]
        tokens_embeddings = tokens_embeddings + tokens_embeddings_iter

    if method == 'Elo':
        final_results = []
        tokens_input_gen_iter = 0
        tokens_output_gen_iter = 0
        tokens_input_test_iter = 0
        tokens_output_test_iter = 0
        while iteration > 0:
            print(f"Iteration number {number_of_iteration}")
            prompt_contents = [item['prompt'] for item in old_prompts]
            combine_prompts = []
            new_results_prompts_cost = iterations(test_cases, method, prompt_contents, number_of_prompts - best_prompts, None, model_test, model_test_temperature, model_test_max_tokens, model_iteration, model_iteration_temperature, model_iteration_max_tokens, None, description, best_prompts, timeout, n_retries)
            final_results.append(new_results_prompts_cost)
            new_results = new_results_prompts_cost[0]
            cost = cost + new_results_prompts_cost[1]
            tokens_input_gen_iter = tokens_input_gen_iter + new_results_prompts_cost[2]
            tokens_output_gen_iter = tokens_output_gen_iter + new_results_prompts_cost[4]
            tokens_input_test_iter = tokens_input_test_iter + new_results_prompts_cost[3]
            tokens_output_test_iter = tokens_output_test_iter + new_results_prompts_cost[5]

            iteration = iteration - 1
            number_of_iteration = number_of_iteration + 1
            old_prompts = new_results[1]
    
    return final_results
